import sys
import yaml
import argparse
import numpy as np
import pylab as pl
import corner

from astropy.io import ascii

from numpy.lib.recfunctions import append_fields, repack_fields

from speedyfit import mcmc, model, plotting, fileio, filters, photometry_query
from speedyfit.default_setup import default_double, default_single



def select_photometry(photbands, obs, obs_err, remove_nan=True, remove_color=True, include=None, exclude=None,
                      verbose=True):
    """
    Function to select the wanted photometry.

    - removes photometry with nan values in value or error
    - removed colors
    - selects photometry based on include and exclude arrays
    """

    # -- remove photometry with nan values in measurement or error.
    if remove_nan:
        nani = np.isnan(obs) | np.isnan(obs_err)
        if any(nani) and verbose:
            print("Warning: there are NaN values in the following photometric bands:")
            for p in photbands[nani]:
                print("\t {}".format(p))
            print("They were removed")
        obs, obs_err, photbands = obs[~nani], obs_err[~nani], photbands[~nani]

    # -- remove colors
    if remove_color:
        color = np.array([filters.is_color(p) for p in photbands])
        s = np.where(~color)
        if any(color) and verbose:
            print("Warning: The following colors were removed:")
            for p in photbands[s]:
                print("\t {}".format(p))
        photbands, obs, obs_err = photbands[s], obs[s], obs_err[s]

    # -- only include bands that are requested based on the include keyword, or exclude all non wanted photometry
    #    based on the exclude keyword.
    if include is not None:
        photsys_include = [x for x in include if not '.' in x]
        photband_include = [x for x in include if '.' in x]
        incband = []
        for i, photband in enumerate(photbands):
            if photband in photband_include or photband.split('.')[0] in photsys_include:
                incband.append(i)
        incband = (np.array(incband),)
        photbands, obs, obs_err = photbands[incband], obs[incband], obs_err[incband]

    elif exclude is not None:
        photsys_exclude = [x for x in exclude if not '.' in x]
        photband_exclude = [x for x in exclude if '.' in x]
        incband = []
        for i, photband in enumerate(photbands):
            if photband not in photband_exclude and photband.split('.')[0] not in photsys_exclude:
                incband.append(i)
        incband = (np.array(incband),)
        photbands, obs, obs_err = photbands[incband], obs[incband], obs_err[incband]

    return photbands, obs, obs_err


def get_observations(setup):
    # -- parse photometry
    if isinstance(setup['photband_index'], str):
        data = ascii.read(setup['photometryfile'], format='fixed_width')
        photbands = np.array(data['band'])
        obs = np.array(data['flux'])
        obs_err = np.array(data['eflux'])
    else:
        data = ascii.read(setup['photometryfile'], data_start=0, header_start=None)

        setup['photband_index'] = data.colnames[setup['photband_index']]
        setup['obs_index'] = data.colnames[setup['obs_index']]
        setup['err_index'] = data.colnames[setup['err_index']]

        photbands = np.array(data[setup['photband_index']])
        obs = np.array(data[setup['obs_index']])
        obs_err = np.array(data[setup['err_index']])

    # -- select the requested photometry
    photbands, obs, obs_err = select_photometry(photbands, obs, obs_err, remove_nan=True, remove_color=True,
                                                include=setup.get('photband_include', None),
                                                exclude=setup.get('photband_exclude', None))

    return photbands, obs, obs_err


def fit_sed(setup, photbands, obs, obs_err):

    # -- pars limits
    pnames = setup['pnames']
    limits = np.array(setup['limits'])

    # -- pars constraints
    constraints = setup['constraints']
    for con, val in list(constraints.items()):
        if len(val) == 2:
            constraints[con] = [val[0], val[1], val[1]]

    if 'parallax' in constraints:
        p, pm, pp = constraints.pop('parallax')
        constraints['distance'] = [1000. / p, 1000. * pm / p ** 2, 1000. * pp / p ** 2]

    print("Applied constraints: ")
    for con, val in list(constraints.items()):
        print("\t {} = {} - {} + {}".format(con, val[0], val[1], val[2]))

    if 'distance' in constraints:
        # convert pc to Rsol
        constraints['distance'] = [44365810.04823812 * constraints['distance'][0],
                                   44365810.04823812 * constraints['distance'][1],
                                   44365810.04823812 * constraints['distance'][2], ]

    # -- pars limits on derived properties
    derived_limits = setup['derived_limits']

    # -- pars grid
    gridnames = setup['grids']
    grids = model.load_grids(gridnames, pnames, limits, photbands)

    # -- switch logg to g for a binary system
    if 'q' in constraints:
        if 'logg' in pnames:
            limits[pnames.index('logg')] = 10 ** limits[pnames.index('logg')]
            pnames[pnames.index('logg')] = 'g'

        if 'logg2' in pnames:
            limits[pnames.index('logg2')] = 10 ** limits[pnames.index('logg2')]
            pnames[pnames.index('logg2')] = 'g2'

        if 'logg' in constraints:
            g = 10 ** constraints['logg'][0]
            g_el = g * constraints['logg'][1] * np.log(10)
            g_eu = g * constraints['logg'][2] * np.log(10)
            constraints['g'] = [g, g_el, g_eu]

        if 'logg2' in constraints:
            g = 10 ** constraints['logg2'][0]
            g_el = g * constraints['logg2'][1] * np.log(10)
            g_eu = g * constraints['logg2'][2] * np.log(10)
            constraints['g2'] = [g, g_el, g_eu]

    # -- check for variables that are kept fixed
    fixed = np.where(limits[:, 0] == limits[:, 1])
    varied = np.where(limits[:, 0] != limits[:, 1])

    pnames = np.array(pnames)
    fixed_variables = {}
    for par, val in zip(pnames[fixed], limits[:, 0][fixed]):
        fixed_variables[par] = val

    pnames = list(pnames[varied])
    limits = limits[varied]

    # -- pars mcmc setup
    nwalkers = setup.get('nwalkers', 100)
    nsteps = setup.get('nsteps', 2000)
    nrelax = setup.get('nrelax', 500)
    a = setup.get('a', 10)

    # -- MCMC
    results, samples = mcmc.MCMC(obs, obs_err, photbands,
                                 pnames, limits, grids,
                                 fixed_variables=fixed_variables,
                                 constraints=constraints, derived_limits=derived_limits,
                                 nwalkers=nwalkers, nsteps=nsteps, nrelax=nrelax,
                                 a=a)

    # -- add fixed variables to results dictionary
    for par, val in list(fixed_variables.items()):
        results[par] = [val, val, 0, 0]

    # -- deal with the switch back to logg
    if 'g' in samples.dtype.names:
        samples = append_fields(samples, 'logg', np.log10(samples['g']), usemask=False)

    if 'g' in results:
        results['logg'] = np.log10(results['g'])
        results.pop('g')

    if 'g2' in samples.dtype.names:
        samples = append_fields(samples, 'logg2', np.log10(samples['g2']), usemask=False)

    if 'g2' in results:
        results['logg2'] = np.log10(results['g2'])
        results.pop('g2')

    _ = constraints.pop('g', None)
    _ = constraints.pop('g2', None)

    names = list(samples.dtype.names)
    if 'g' in names: names.remove('g')
    if 'g2' in names: names.remove('g2')
    samples = samples[names]

    percentiles = setup.get('percentiles', [16, 50, 84])
    pc = np.percentile(samples.view(np.float64).reshape(samples.shape + (-1,)), percentiles, axis=0)
    pars = {}
    for p, v, e1, e2 in zip(samples.dtype.names, pc[1], pc[1] - pc[0], pc[2] - pc[1]):
        results[p] = [results[p], v, e1, e2]
        pars[p] = v

    return results, samples, constraints, gridnames


def write_results(setup, results, samples, obs, obs_err, photbands):

    outpars, outvals = [], []
    for par in ['teff', 'logg', 'L', 'rad']:
        outpars.append(par)
        outpars.append(par + '_err')
        outvals.append(results[par][1])
        outvals.append(np.average([results[par][2], results[par][3]]))

    if 'teff2' in results:
        for par in ['teff2', 'logg2', 'L2', 'rad2']:
            outpars.append(par)
            outpars.append(par + '_err')
            outvals.append(results[par][1])
            outvals.append(np.average([results[par][2], results[par][3]]))

    resultfile = setup.get('resultfile', None)
    if resultfile is not None:
        import pandas as pd
        data = pd.DataFrame(data=[outvals], columns=outpars)
        data.to_csv(resultfile, index=False)

    datafile = setup.get('datafile', None)
    if not datafile is None:
        # -- get plain text settings to retain comments and ordering
        setupfile = open(args.filename)
        setup_str = "".join(setupfile.readlines())
        setupfile.close()

        fileio.write2fits(samples, datafile, setup=setup_str)

    h5file = setup.get('h5file', None)
    if h5file is not None:
        fileio.write_summary2hdf5(setup['objectname'], samples, obs, obs_err, photbands, pars=results,
                                  grids=setup['grids'], filename=h5file)


def plot_results(setup, results, samples, constraints, gridnames, obs, obs_err, photbands):

    # check for 10 possible plots. Should be enough for now.
    for i in range(10):

        pindex = 'plot' + str(i)
        if not pindex in setup: continue

        if setup[pindex]['type'] == 'sed_fit':

            res = setup[pindex].get('result', 'best')

            pl.figure(i)
            pl.subplots_adjust(wspace=0.25)
            plotting.plot_fit(obs, obs_err, photbands, pars=results, constraints=constraints, grids=setup['grids'],
                              gridnames=gridnames, result=res)

            if not setup[pindex].get('path', None) is None:
                pl.savefig(setup[pindex].get('path', 'sed_fit.png'))

        if setup[pindex]['type'] == 'constraints':

            pl.figure(i, figsize=(2 * len(constraints), 6))
            pl.subplots_adjust(wspace=0.40, left=0.07, right=0.98)

            plotting.plot_constraints(constraints, samples, results)

            if not setup[pindex].get('path', None) is None:
                pl.savefig(setup[pindex].get('path', 'constraints.png'))

        if setup[pindex]['type'] == 'distribution':

            pars1 = []
            for p in setup[pindex].get('parameters', ['teff', 'rad', 'L', 'd']):
                if p in samples.dtype.names: pars1.append(p)

            data = repack_fields(samples[pars1])

            if setup[pindex].get('show_best', False):
                truths = [results[p][0] for p in data.dtype.names]
            else:
                truths = None

            fig = corner.corner(data.view(np.float64).reshape(data.shape + (-1,)),
                                labels=data.dtype.names,
                                quantiles=setup[pindex].get('quantiles', [0.025, 0.16, 0.5, 0.84, 0.975]),
                                levels=setup[pindex].get('levels', [0.393, 0.865, 0.95]),
                                truths=truths,
                                show_titles=True, title_kwargs={"fontsize": 12}, )

            if not setup[pindex].get('path', None) is None:
                pl.savefig(setup[pindex].get('path', 'distribution.png'))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', action="store", type=str, help='use setup given in this file')
    parser.add_argument("-empty", dest='empty', type=str, default=None,
                        help="Create empty setup file ('single' or 'double')")
    parser.add_argument('--phot', dest='photometry', action='store_true',
                        help='When creating a new setupfile, use this option to also download photometry from Vizier and Tap archives.')
    parser.add_argument('--noplot', dest='noplot', action='store_true',
                        help="Don't show any plots, only store to disk.")
    args, variables = parser.parse_known_args()

    if args.empty is not None:

        objectname = args.filename
        filename = objectname + '_single.yaml' if args.empty == 'single' else objectname + '_binary.yaml'

        plx, e_plx = photometry_query.get_parallax(objectname)

        out = default_single if args.empty == 'single' else default_double
        out = out.replace('<photfilename>', objectname + '.phot')
        out = out.replace('<objectname>', objectname)
        out = out.replace('<plx>', str(plx))
        out = out.replace('<e_plx>', str(e_plx))

        ofile = open(filename, 'w')
        ofile.write(out)
        ofile.close()

        if args.photometry:
            photometry = photometry_query.get_photometry(objectname, filename=objectname + '.phot')

        sys.exit()

    if args.filename is None:
        print("Nothing to do")
        sys.exit()

    # -- load the setup file
    setupfile = open(args.filename)
    setup = yaml.safe_load(setupfile)
    setupfile.close()

    # -- obtain the observations
    photbands, obs, obs_err = get_observations(setup)

    # -- perform the SED fit
    results, samples, constraints, gridnames = fit_sed(setup, photbands, obs, obs_err)

    # -- write the results
    write_results(setup, results, samples, obs, obs_err, photbands)

    # -- create plots
    plot_results(setup, results, samples, constraints, gridnames, obs, obs_err, photbands)

    print("================================================================================")
    print("")
    print("Resulting parameter values and errors:")
    print("   Par             Best        Pc       emin       emax")
    for p in samples.dtype.names:
        print("   {:10s} = {}   {}   -{}   +{}".format(p, *plotting.format_parameter(p, results[p])))

    if not args.noplot:
        pl.show()


if __name__ == "__main__":
    main()



import copy
import os
import argparse

import yaml

import pandas as pd

from astropy.coordinates import Angle

from speedyfit import photometry_query
from speedyfit.speedyfit import fit_sed, get_observations, write_results, plot_results
from speedyfit.default_setup import default_double, default_single


def process_objects(data):

    def convert_ra(ra):
        if ' ' in str(ra).strip() or ':' in str(ra).strip():
            ra = str(ra).strip()
            ra = ra.replace(' ', '').replace(':', '')
        else:
            a = Angle(ra, unit='degree').hms
            ra = '{:02.0f}{:02.0f}{:05.2f}'.format(*a)
        return ra

    def convert_dec(dec):
        if ' ' in str(dec).strip() or ':' in str(dec).strip():
            dec = str(dec).strip()
            dec = dec.replace(' ', '').replace(':', '')
            if not '-' in dec and not '+' in dec:
                dec = '+' + dec
        else:
            a = Angle(dec, unit='degree').dms
            dec = '{:+03.0f}{:02.0f}{:05.2f}'.format(a[0], abs(a[1]), abs(a[2]))
        return dec

    if 'name' in data.columns.values:
        object_list = data['name'].apply(lambda x: x.replace(' ', '_')).values
    else:
        # deal with coordinates
        ra_ = data['ra'].apply(convert_ra)
        dec_ = data['dec'].apply(convert_dec)

        name = ['J{}{}'.format(r, d) for r, d in zip(ra_, dec_)]

        object_list = pd.Series(data=name)

    return object_list


def read_setup(setup):

    if os.path.isfile(setup):
        infile = open(setup, 'r')
        setup = infile.readlines()
        infile.close()
        setup = ''.join(setup)

    elif setup == 'single':
        setup = default_single
    elif setup == 'double':
        setup = default_double
    else:
        print('Setup not recognized! Using single as default')
        setup = default_single

    return setup


def prepare_setup(object_list, basedir, default_setup):

    for objectname in object_list:

        if not os.path.isdir(basedir+'/'+objectname):
            os.mkdir(basedir+'/'+objectname)

        plx, e_plx = photometry_query.get_parallax(objectname)

        out = copy.copy(default_setup)
        if '<photfilename>' in out:
            photfilename_ = basedir + '/' + objectname + '/' + objectname + '.phot'
            out = out.replace('<photfilename>', photfilename_)
        if '<objectname>' in out:
            objectname_ = basedir + '/' + objectname + '/' + objectname
            out = out.replace('<objectname>', objectname_)
        if '<plx>' in out:
            out = out.replace('<plx>', str(plx))
            out = out.replace('<e_plx>', str(e_plx))

        filename = basedir + '/' + objectname + '/' + objectname + '_setup.yaml'

        ofile = open(filename, 'w')
        ofile.write(out)
        ofile.close()


def prepare_photometry(object_list, basedir, skip_existing=True):

    for objectname in object_list:

        if os.path.isfile(basedir+'/'+objectname+'/'+objectname+'.phot') and skip_existing:
            continue

        if not os.path.isdir(basedir+'/'+objectname):
            os.mkdir(basedir+'/'+objectname)

        try:
            filename = basedir + '/' + objectname + '/' + objectname + '.phot'
            photometry = photometry_query.get_photometry(objectname, filename=filename)
        except Exception as e:
            print("Failed to obtain photometry for {}".format(objectname))
            print(e)


def fit_seds(object_list, basedir):

    all_results = {}

    for objectname in object_list:

        # read the setup
        filename = basedir + '/' + objectname + '/' + objectname + '_setup.yaml'
        try:
            infile = open(filename)
            setup = yaml.safe_load(infile)
            infile.close()
        except Exception as e:
            print("Could not read setupfile for {}".format(objectname))
            print(e)
            continue

        try:
            # get the photometry
            photbands, obs, obs_err = get_observations(setup)

            # do the SED fit
            results, samples, constraints, gridnames = fit_sed(setup, photbands, obs, obs_err)

            # add results to results dic
            all_results.setdefault('objectname', []).append(objectname)
            for k, v in results.items():
                all_results.setdefault(k, []).append(v[0])
                all_results.setdefault(k+'_err', []).append(( v[2] + v[3] ) / 2.0)

            # write the results
            write_results(setup, results, samples, obs, obs_err, photbands)

            # make any requested plots
            plot_results(setup, results, samples, constraints, gridnames, obs, obs_err, photbands)

        except Exception as e:
            print("Could not fit {}".format(objectname))
            print(e)

    all_results = pd.DataFrame(data=all_results)
    all_results.to_csv('results.csv')
    print(all_results)



def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', action="store", type=str,
                        help='file containing a list with the names or coordinates of the systems to fit')
    parser.add_argument("-setup", dest='setup', type=str, default='single',
                        help="The path to the setup file to use, or a standard: 'single' or 'double'. "
                             "(default = 'single'")
    parser.add_argument("-basedir", dest='basedir', type=str, default='./',
                        help="The directory where all photometry, setup and results will be stored. (default = ./)")
    parser.add_argument("--fit", dest='fit', action='store_true',
                        help="Fit the systems")
    parser.add_argument('--phot', dest='phot', action='store_true',
                        help='Obtain photometry of the systems')
    parser.add_argument('--noplot', dest='noplot', action='store_true',
                        help="Don't show any plots, only store to disk.")
    args, variables = parser.parse_known_args()

    basedir = args.basedir
    default_setup = read_setup(args.setup)

    if not os.path.isdir(basedir):
        os.mkdir(basedir)

    object_data = pd.read_csv(args.filename)
    object_list = process_objects(object_data)

    # create the setup file for all systems
    prepare_setup(object_list, basedir, default_setup)

    # try to obtain photometry for all systems if requested
    if args.phot:
        prepare_photometry(object_list, basedir)

    # fit all systems if requested
    if args.fit:
        fit_seds(object_list, basedir)


import os
import h5py
import datetime

import numpy as np

try:
   import gzip
   has_zip = True
except Exception as d:
   has_zip = False

from astropy.io import fits

def read2list(filename, commentchar='#', splitchar=None, skip_empty=True, skip_lines=0, **kwargs):
   """
   Load an ASCII file to list of lists.
   
   The comments and data go to two different lists.
   
   Also opens gzipped files.
   
   @param filename: name of file with the data
   @type filename: string
   @keyword commentchar: character(s) denoting comment rules
   @type commentchar: list of str
   @keyword splitchar: character seperating entries in a row (default: whitespace)
   @type splitchar: str or None
   @keyword skip_empty: skip empty lines
   @type skip_empty: bool
   @keyword skip_lines: skip nr of lines (including comment and empty lines)
   @type skip_lines: integer
   @return: list of lists (data rows)
            list of lists (comments lines without commentchar),
   @rtype: (list,list)
   """
   
   if os.path.splitext(filename)[1] == '.gz' and has_zip:
      ff = gzip.open(filename)
   elif os.path.splitext(filename)[1] == '.gz':
      print('gzip not installed')
      return [], []
   else:
      ff = open(filename)
      
   data = []  # data
   comm = []  # comments
   
   line_nr = -1
   
   #-- fixwidth split or character split?
   if splitchar is None or isinstance(splitchar,str):
      fixwidth = False
   else:
      fixwidth = True
   
   while 1:  # might call read several times for a file
      line = ff.readline()
      if not line: break  # end of file
      line_nr += 1
      if line_nr<skip_lines:        
            continue
      
      #-- strip return character from line
      if skip_empty and line.isspace():
            continue # empty line
      
      #-- remove return characters
      line = line.replace('\n','')
      #-- when reading a comment line
      if line[0] in commentchar:
            comm.append(line[1:])
            continue # treat next line
      
      #-- when reading data, split the line
      if not fixwidth:
            data.append(line.split(splitchar))
      else:
            data.append(fw2python(line,splitchar))
   ff.close()
   
   #-- and return the contents
   return data,comm

def read2array(filename,**kwargs):
   """
   Load ASCII file to a numpy array.
   
   For a list of extra keyword arguments, see <read2list>.
   
   If you want to return a list of the columns instead of rows, just do
   
   >>> col1,col2,col3 = ascii.read2array(myfile).T
   
   @param filename: name of file with the data
   @type filename: string
   @keyword dtype: type of numpy array (default: float)
   @type dtype: numpy dtype
   @keyword return_comments: flag to return comments (default: False)
   @type return_comments: bool
   @return: data array (, list of comments)
   @rtype: ndarray (, list)
   """
   dtype = kwargs.get('dtype',np.float)
   return_comments = kwargs.get('return_comments',False)
   data,comm = read2list(filename,**kwargs)
   data = np.array(data,dtype=dtype)
   return return_comments and (data,comm) or data

def write_array(data, filename, **kwargs):
    """
    Save a numpy array to an ASCII file.
    
    Add comments via keyword comments (a list of strings denoting every comment
    line). By default, the comment lines will be preceded by the C{commentchar}.
    If you want to override this behaviour, set C{commentchar=''}.
    
    If you give a record array, you can simply set C{header} to C{True} to write
    the header, instead of specifying a list of strings.
    
    @keyword header: optional header for column names
    @type header: list of str (or boolean for record arrays)
    @keyword comments: comment lines
    @type comments: list of str
    @keyword commentchar: comment character
    @type commentchar: str
    @keyword sep: separator for the columns and header names
    @type sep: str
    @keyword axis0: string denoting the orientation of the matrix. If you gave
    a list of columns, set C{axis0='cols'}, otherwise C{axis='rows'} (default).
    @type axis0: str, one of C{cols}, C{rows}.
    @keyword mode: file mode (a for appending, w for (over)writing...)
    @type mode: char (one of 'a','w'...)
    @keyword auto_width: automatically determine the width of the columns
    @type auto_width: bool
    @keyword formats: formats to use to write each column
    @type formats: list of string formatters
    """
    header = kwargs.get('header',None)
    comments = kwargs.get('comments',None)
    commentchar = kwargs.get('commentchar','#')
    sep = kwargs.get('sep',' ')
    axis0 = kwargs.get('axis0','rows')
    mode = kwargs.get('mode','w')
    auto_width = kwargs.get('auto_width',False)
    formats = kwargs.get('formats',None)
    # use '%g' or '%f' or '%e' for writing floats automatically from record arrays with auto width
    use_float = kwargs.get('use_float','%f') 
    
    #-- switch to rows first if a list of columns is given
    if not isinstance(data,np.ndarray):
        data = np.array(data)
    if not 'row' in axis0.lower():
        data = data.T
    
    if formats is None:
        try:
            formats = [('S' in str(data[col].dtype) and '%s' or use_float) for col in data.dtype.names]
        except TypeError:
            formats = [('S' in str(col.dtype) and '%s' or '%s') for col in data.T]
    #-- determine width of columns: also take the header label into account
    col_widths = []
    #-- for record arrays
    if auto_width is True and header==True:
        for fmt,head in zip(formats,data.dtype.names):
            col_widths.append(max([len('%s'%(fmt)%(el)) for el in data[head]]+[len(head)]))
    #-- for normal arrays and specified header
    elif auto_width is True and header is not None:
        for i,head in enumerate(header):
            col_widths.append(max([len('%s'%(formats[i])%(el)) for el in data[:,i]]+[len(head)]))
    #-- for normal arrays without header
    elif auto_width is True and header is not None:
        for i in range(data.shape[1]):
            col_widths.append(max([len('%s'%(formats[i])%(el)) for el in data[:,i]]))
    
    if header is True:
        col_fmts = [str(data.dtype[i]) for i in range(len(data.dtype))]
        header = data.dtype.names
    else:
        col_fmts = None
    
    ff = open(filename,mode)
    if comments is not None:
        ff.write('\n'.join(comments)+'\n')
    
    #-- WRITE HEADER
    #-- when header is desired and automatic width
    if header is not None and col_widths:
        ff.write('#'+sep.join(['%%%s%ss'%(('s' in fmt and '-' or ''),cw)%(head) for fmt,head,cw in zip(formats,header,col_widths)])+'\n')
    #-- when header is desired
    elif header is not None:
        ff.write('#'+sep.join(header)+'\n')
    
    #-- WRITE COLUMN FORMATS
    if col_fmts is not None and col_widths:
        ff.write('#'+sep.join(['%%%s%ss'%(('s' in fmt and '-' or ''),cw)%(colfmt) for fmt,colfmt,cw in zip(formats,col_fmts,col_widths)])+'\n')
    elif col_fmts is not None:
        ff.write('#'+sep.join(['%%%ss'%('s' in fmt and '-' or '')%(colfmt) for fmt,colfmt in zip(formats,col_fmts)])+'\n')
    
    #-- WRITE DATA
    #-- with automatic width
    if col_widths:
        for row in data:
            ff.write(' '+sep.join(['%%%s%s%s'%(('s' in fmt and '-' or ''),cw,fmt[1:])%(col) for col,cw,fmt in zip(row,col_widths,formats)])+'\n')
    #-- without automatic width
    else:
        for row in data:
            ff.write(sep.join(['%s'%(col) for col in row])+'\n')
    ff.close()

def write2hdf5(samples, blobs, filename):
   """
   
   """
   
   hdf = h5py.File(filename)
   
   hdf.create_dataset('samples', data=samples)
   hdf.create_dataset('blobs', data=blobs)
   
   hdf.close()
   
   
def read_hdf5(filename):
   
   hdf = h5py.File(filename, 'r')
   
   samples = hdf['samples'].value
   blobs = hdf['blobs'].value
   
   hdf.close()
   
   return samples, blobs
   
   
def write2fits(samples, filename, setup=None):
   """
   writes recarray containing all samples to fits file
   """
   
   #-- create Table data for fits file
   cols = []
   
   for (name, fmt) in samples.dtype.descr:
      col = fits.Column(name=name, format='D', array=samples[name])
      cols.append(col)
      
   cols = fits.ColDefs(cols)
   
   tbhdu = fits.BinTableHDU.from_columns(cols)
   
   #-- add date to header
   header = tbhdu.header
   header['date'] = str(datetime.datetime.now())
   
   #-- add complete setup as comment if given as str
   for line in setup.split('\n'):
      header['comment'] = line
   
   #-- delete file if it already exists
   if os.path.isfile(filename):
      os.remove(filename)
   
   tbhdu.writeto(filename)
   
def read_fits(filename):
   """
   Reads fits file and returns samples and the settings if they exist
   """
   
   hdu = fits.open(filename)
   
   samples = hdu[1].data
   
   setup = hdu[1].header['comment']
   
   return samples, str(setup)
   

def write_summary2hdf5(objectname, samples, obs, obs_err, photbands, pars={}, grids=[], filename=None):
   """
   Writes the obeservations and results to hdf5 format readable by AOTS
   """
   from astropy.table import Table
   from . import model
   from . import filters
   from .photometry_query import get_coordinate
   
   if filename is None: filename = objectname + '_sedfit.hdf5'
   f = h5py.File(filename, "w")
   
   ra, dec = get_coordinate(objectname)
      
   f.attrs['systemname'] = objectname
   f.attrs['ra'] = ra
   f.attrs['dec'] = dec
   
   f.attrs['name'] = 'SED fit'
   f.attrs['note'] = ''
   f.attrs['type'] = 'sedfit'
   
   pars = pars.copy()
   
   # use model from 'best' results or 'pc' results
   result = 'best'
   resi = 0 if result == 'best' else 1
   
   
   colors = np.array([filters.is_color(p) for p in photbands])
   waves = np.array([filters.eff_wave(p) for p in photbands[~colors]])
   pbands = np.array(photbands[~colors], dtype='<S') # needed to convert to S type for compatibility with h5 format.
   obsdata = Table([waves, pbands, obs[~colors], obs_err[~colors]], names=['wave', 'photband', 'flux', 'flux_err'])
   
   #-- create the data group
   data = f.create_group('DATA')
   data.attrs['xlabel'] = 'Wavelength (AA)'
   data.attrs['ylabel'] = 'flux'
   data.attrs['xmin'] = np.min(waves)
   data.attrs['xmax'] = np.max(waves)
   data.attrs['xscale'] = 'log'
   data.attrs['yscale'] = 'log'
   
   obsgr = data.create_dataset('Obs', data=obsdata)
   obsgr.attrs['label'] = 'OBS'
   obsgr.attrs['datatype'] = 'discrete'
   obsgr.attrs['xpar'] = 'wave'
   obsgr.attrs['ypar'] = 'flux'
   
   
   def get_unit(par):
      
      if 'teff' in par: return 'K'
      if 'logg' in par: return 'dex'
      if 'L' in par: return 'Lsol'
      if 'rad' in par: return 'Rsol'
      if 'd' in par: return 'pc'
      if 'ebv' in par: return ''
   
      return ''
   
   if not len(list(pars.keys())) == 0:
      
      
      #-- create parameters
      group = f.create_group('PARAMETERS')
      ipars = pars.copy()
      for key, value in list(pars.items()):
         
         ipars[key] = [value[resi]]
         pars[key] = value[resi]
         
         # if key == 'scale' or key == 'chi2' or key == 'mass' or key == 'logg': continue
         if key in ['scale', 'chi2', 'mass', 'logg']: continue
         if key not in samples.dtype.names: continue
         
         pg = group.create_group(key)
         pg.attrs['unit'] = get_unit(key)
         pg.attrs['value'] = value[resi]
         pg.attrs['err'] = value[2]
         if key == 'g' or key == 'logg':
            pg.attrs['valid'] = False
         else:
            pg.attrs['valid'] = True
         
         y, x = np.histogram(samples[key], 20, density=False)
         x = (x[0:-1] + x[1:])/2.
         y = y / np.float(np.max(y))
         
         dis = Table([x, y], names=['x', 'y'])
         
         
         disgr = pg.create_dataset('DISTRIBUTION', data=dis)
         disgr.attrs['datatype'] = 'bars'
         disgr.attrs['xpar'] = 'x'
         disgr.attrs['ypar'] = 'y'
         
      _ = ipars.pop('d')
      _ = pars.pop('d')
      
      
       #-- create the model group
      group = f.create_group('MODEL')
      group.attrs['xlabel'] = 'Wavelength (AA)'
      group.attrs['ylabel'] = 'Flux'
      #group.attrs['xmin'] = 0.
      #group.attrs['xmax'] = 1.
      
      
      syn, Labs = model.get_itable(grid=grids, photbands=photbands, **ipars)
      syn = syn[:,0]
      
      scale = pars['scale']
      
      
      itable = Table([waves, scale*syn], names=['wave', 'flux'])
      
      itb = group.create_dataset('Iflux', data=itable)
      itb.attrs['label'] = 'Synth. photometry'
      itb.attrs['datatype'] = 'discrete'
      itb.attrs['xpar'] = 'wave'
      itb.attrs['ypar'] = 'flux'
      
      
      #-- if possible, plot a non integrated model
      #   can only be done if provided gridnames are not paths to integrated files.
      if len(grids) > 0 and not os.path.isfile(grids[0]):
         #-- synthetic model
         wave, flux = model.get_table(grid=grids, **pars)
         s = np.where((wave > np.min(waves)-1000) & (wave < np.max(waves) + 1000))
         table = Table([wave[s], scale*flux[s]], names=['wave', 'flux'])
         
         name = 'binary' if 'teff2' in pars else grids[0]
         tb = group.create_dataset(name, data=table)
         tb.attrs['label'] = name
         tb.attrs['datatype'] = 'continuous'
         tb.attrs['xpar'] = 'wave'
         tb.attrs['ypar'] = 'flux'
         
         
         #-- plot components
         if 'teff2' in pars:
            wave, flux = model.get_table(grid=grids[0], teff=pars['teff'], logg=pars['logg'], rad=pars['rad'], ebv=pars['ebv'])
            s = np.where((wave > np.min(waves)-1000) & (wave < np.max(waves) + 1000))
            table1 = Table([wave[s], scale*flux[s]], names=['wave', 'flux'])
            
            tb1 = group.create_dataset(grids[0], data=table1)
            tb1.attrs['label'] = grids[0]
            tb1.attrs['datatype'] = 'continuous'
            tb1.attrs['xpar'] = 'wave'
            tb1.attrs['ypar'] = 'flux'
            
            wave, flux = model.get_table(grid=grids[1], teff=pars['teff2'], logg=pars['logg2'], rad=pars['rad2'], ebv=pars['ebv'])
            s = np.where((wave > np.min(waves)-1000) & (wave < np.max(waves) + 1000))
            table2 = Table([wave[s], scale*flux[s]], names=['wave', 'flux'])
            
            tb2 = group.create_dataset(grids[1], data=table2)
            tb2.attrs['label'] = grids[1]
            tb2.attrs['datatype'] = 'continuous'
            tb2.attrs['xpar'] = 'wave'
            tb2.attrs['ypar'] = 'flux'
            
            
         #-- create the O-C data
         group = f.create_group('O-C')
         group.attrs['xlabel'] = 'Wavelength (AA)'
         group.attrs['ylabel'] = 'O-C (mag)'
         group.attrs['xscale'] = 'log'
         group.attrs['xmin'] = np.min(waves)-1000
         group.attrs['xmax'] = np.max(waves)+1000
         
         
         y = 2.5*np.log10(syn*scale / obs[~colors])
         yerr = 2.5 / np.log(10.0) * obs_err[~colors] / obs[~colors]
         obsdata = Table([waves, y, yerr], names=['wave', 'o-c', 'o-c_err'])
         
         octb = group.create_dataset('Obs', data=obsdata)
         octb.attrs['label'] = 'Obs - Calc'
         octb.attrs['datatype'] = 'discrete'
         octb.attrs['xpar'] = 'wave'
         octb.attrs['ypar'] = 'o-c'
               
   f.close()



import pytest

import numpy as np

from speedyfit import model

import os


class TestModel:

    def test_itable_single(self):

        print('OS: ' , os.environ.get('SPEEDYFIT_MODELS', None))

        photbands = ['STROMGREN.U', 'STROMGREN.B']
        gridfilename = 'kurucz2'

        axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
                                                                           teffrange=(4000, 5000), loggrange=(3.5, 4.0),
                                                                           ebvrange=(0.0, 0.02),
                                                                           variables=['teff', 'logg', 'ebv'])

        grid = [axis_values, pixelgrid]

        # -- check if interpolated model is correct
        flux1, Labs1 = model.get_itable_single(teff=4750, logg=3.75, ebv=0.01, grid=grid)

        assert Labs1 == pytest.approx(0.45605564, 0.001), \
            'Basic model: Luminosity is not correct ({} != 0.456)'.format(Labs1)

        # -- check if radius is treated correctly
        flux1, Labs1 = model.get_itable_single(teff=4750, logg=3.75, ebv=0.01, rad=1, grid=grid)
        assert Labs1 == pytest.approx(0.45605564, 0.001), \
            'Solar Radius: Luminosity is not correct ({} != 0.456)'.format(Labs1)

        flux1, Labs1 = model.get_itable_single(teff=4750, logg=3.75, ebv=0.01, rad=2, grid=grid)
        assert Labs1 == pytest.approx(2 ** 2 * 0.45605564, 0.001), \
            'Double Radius: Luminosity is not correct ({} != 1.824)'.format(Labs1)

        # -- compare with get_itable
        flux1, Labs1 = model.get_itable_single(teff=4750, logg=3.75, ebv=0.01, grid=grid)
        flux2, Labs2 = model.get_itable(teff=4750, logg=3.75, ebv=0.01, grid=grid)

        for f1, f2 in zip(flux1, flux2):
            assert f1 == f2, \
                "flux output of get_itable should be same as get_itable_single, ({} != {})".format(flux1, flux2)

        assert Labs1[0] == Labs2['L'], \
            "Labs output of get_itable should be same as get_itable_single, ({} != {})".format(Labs1[0], Labs2['L'])

    def test_itable(self):

        photbands = ['STROMGREN.U', 'STROMGREN.B']
        gridfilename = 'kurucz2'

        axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
                                                                           teffrange=(4000, 5000), loggrange=(3.5, 4.0),
                                                                           ebvrange=(0.0, 0.02),
                                                                           variables=['teff', 'logg', 'ebv'])

        grid1 = [axis_values, pixelgrid]

        gridfilename = 'tmap'

        axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
                                                                           teffrange=(20000, 40000),
                                                                           loggrange=(5.0, 6.5),
                                                                           ebvrange=(0.0, 0.02),
                                                                           variables=['teff', 'logg', 'ebv'])

        grid2 = [axis_values, pixelgrid]

        flux1, Labs1 = model.get_itable(teff=4750, logg=3.75, ebv=0.01, rad=1, grid=grid1)
        flux2, Labs2 = model.get_itable(teff=4250, logg=3.50, ebv=0.01, rad=1, grid=grid1)

        fluxt, Labst = model.get_itable(teff=4750, logg=3.75, ebv=0.01, rad=1,
                                        teff2=4250, logg2=3.50, ebv2=0.01, rad2=1, grid=[grid1, grid1])

        for f1, f2, ft in zip(flux1, flux2, fluxt):
            assert f1 + f2 == ft, \
                "binary SED should have sum of flux of single SED, ({} + {} != {})".format(flux1, flux2, fluxt)

        flux1, Labs1 = model.get_itable(teff=4800, logg=3.8, ebv=0.01, rad=2, grid=grid1)
        flux2, Labs2 = model.get_itable(teff=25000, logg=5.8, ebv=0.01, rad=0.14, grid=grid2)

        fluxt, Labst = model.get_itable(teff=4800, logg=3.8, ebv=0.01, rad=2,
                                        teff2=25000, logg2=5.8, ebv2=0.01, rad2=0.14,
                                        grid=[grid1, grid2])

        for f1, f2, ft in zip(flux1, flux2, fluxt):
            assert f1 + f2 == ft, \
                "binary SED should have sum of flux of single SED, ({} + {} != {})".format(flux1, flux2, fluxt)

    def test_itable_vega(self):
        """
        Test the returned flux for Vega
        Teff = 9602 +- 180 K
        logg = 4.10 +- 0.1
        Rad = 2.362 Rsol
        d = 7.68 +- 0.02 pc
        """

        photbands = ['STROMGREN.U', 'STROMGREN.B', 'STROMGREN.V', 'STROMGREN.Y', '2MASS.KS']
        gridfilename = 'kurucz2'

        axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
                                                                           teffrange=(8000, 10000),
                                                                           loggrange=(4.0, 5.0),
                                                                           ebvrange=(0.0, 0.02),
                                                                           variables=['teff', 'logg', 'ebv'])

        grid = [axis_values, pixelgrid]

        syn, Labs = model.get_itable_single(teff=9602, logg=4.10, ebv=0.0, rad=2.362, grid=grid)

        syn = syn / 340745610.888946 ** 2  # correct for distance
        Labs = Labs[0]

        obs = np.array([8.528e-10, 5.664e-09, 6.190e-09, 3.547e-09, 3.803e-11])
        obs_e = np.array([3.458e-11, 1.484e-10, 1.788e-10, 9.148e-11, 6.515e-12])

        ratio = (obs / syn)
        weights = (obs / obs_e)
        scale = np.average(ratio, weights=weights)
        e_scale = np.sqrt(np.dot(weights, (ratio - scale) ** 2) / weights.sum())

        assert abs(scale - 1.0) < e_scale, \
            "Cannot reproduce VEGA at correct distance! scale = {:0.1f}+-{:0.1f} instead of 1".format(scale, e_scale)

        assert abs(Labs - 40) < 2.5, "Vega should have luminosity of 40.12 Lsol, model gives {:0.2f} ".format(Labs)

    def test_itable_EG50(self):
        """
        Test the returned flux for a known white dwarf star EG 50
        Teff = 22290 +- 350 K
        logg = 8.10 +- 0.05
        Rad = 0.012 +- 0.001 Rsol
        d = 19 +- 3 pc
        """

        photbands = ['STROMGREN.U', 'STROMGREN.V', 'STROMGREN.B', 'STROMGREN.Y', '2MASS.H']
        gridfilename = 'tmap'

        axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
                                                                           teffrange=(20000, 40000),
                                                                           loggrange=(5.0, 6.5),
                                                                           ebvrange=(0.0, 0.02),
                                                                           variables=['teff', 'logg', 'ebv'])

        grid = [axis_values, pixelgrid]

        syn, Labs = model.get_itable_single(teff=22290, logg=6.5, ebv=0.0, rad=0.012, grid=grid)

        syn = syn / 842950390.9165244 ** 2  # correct for distance

        obs = np.array([4.89e-14, 1.02e-13, 9.31e-14, 5.34e-14, 9.76e-16])
        obs_e = np.array([2.53e-15, 3.99e-15, 3.08e-15, 1.72e-15, 5.48e-17])

        ratio = (obs / syn)
        weights = (obs / obs_e)
        scale = np.average(ratio, weights=weights)
        e_scale = np.sqrt(np.dot(weights, (ratio - scale) ** 2) / weights.sum())

        assert abs(scale - 1.0) < e_scale, \
            "Cannot reproduce EG 50 at correct distance! scale = {:0.1f}+-{:0.1f} instead of 1".format(scale, e_scale)



import pytest

import pandas as pd

from speedyfit.speedyfit_batch import process_objects

class TestBatch:

    def test_process_objects(self):

        object_data = pd.DataFrame(data={
            'name':['GALEXJ153213.7+010451', 'HD42709', 'Gaia DR2 1209876314302933632'],
            'info':['', 'intersting star', 'bla']
        })

        object_list = process_objects(object_data)

        assert len(object_list) == 3
        assert object_list[2] == 'Gaia_DR2_1209876314302933632'

        object_data = pd.DataFrame(data={
            'ra': [233.057210996, 10.7429509085],
            'dec': [1.0807152108, -38.1270271017]
        })

        object_list = process_objects(object_data)

        assert len(object_list) == 2
        assert object_list[0] == 'J153213.73+010450.57'
        assert object_list[1] == 'J004258.31-380737.30'


        object_data = pd.DataFrame(data={
            'ra': ['15:32:13.73', '00:42:58.31'],
            'dec': ['+01:04:50.57', '-38:07:37.30']
        })

        object_list = process_objects(object_data)

        assert len(object_list) == 2
        assert object_list[0] == 'J153213.73+010450.57'
        assert object_list[1] == 'J004258.31-380737.30'

        object_data = pd.DataFrame(data={
            'ra': ['15 32 13.73', '00 42 58.31', '15 32 13.73'],
            'dec': ['+01 04 50.57', '-38 07 37.30', '01 04 50.57']
        })

        object_list = process_objects(object_data)

        assert len(object_list) == 3
        assert object_list[0] == 'J153213.73+010450.57'
        assert object_list[1] == 'J004258.31-380737.30'
        assert object_list[2] == 'J153213.73+010450.57'




import os 

import numpy as np
import pylab as pl

import matplotlib.cm as cm
import matplotlib.patches as patches
import matplotlib.ticker as ticker

from scipy.stats import gaussian_kde

from . import model, filters, reddening

from astropy.io import ascii

def format_parameter(name, value):
   
   temp = "{:7.3f}"
   if 'teff' in name or name == 'd':
      temp = "{:7.0f}"
   
   elif  'logg' in name or 'rad' in name or 'mass' in name or name=='L' or name == 'L2':
      temp = "{:7.2f}"
   
   elif 'ebv' in name or 'q' in name:
      temp = "{:7.3f}"
   
   if hasattr(value, '__iter__'):
      return [temp.format(v) for v in value]
   else:
      return temp.format(value)
   
def plot_distribution(data, parameters=None, percentiles=[16, 50, 84]):
   """
   Plots a histogram of the requested parameters, or all parameters if none are given.
   """
   
   if parameters is None:
      parameters = data.dtype.names
   
   x = len(parameters)
   
   for i, par in enumerate(parameters):
      d = data[par]
      
      pc  = np.percentile(d, percentiles, axis=0)
      
      pl.subplot(1,x,i+1)
      
      pl.hist(d, 50, normed=True)
      
      for l in pc:
         pl.axvline(x=l, ls='--', color='k')
         
         emin, emax = pc[1]-pc[0], pc[2]-pc[1]
         
         if np.min([emin, emax]) > 10:
            temp = "{}={:0.0f} -{:0.0f} +{:0.0f}"
         elif np.min([emin, emax]) > 1:
            temp = "{}={:0.1f} -{:0.1f} +{:0.1f}"
         else:
            temp = "{}={:0.2f} -{:0.2f} +{:0.2f}"
         
      pl.title(temp.format(par, pc[1], emin, emax))

def plot_distribution_2d(data, xpar, ypar, percentiles=[16, 50, 84]):
   
   x, y = data[xpar], data[ypar]
   
   xpc  = np.percentile(x, percentiles, axis=0)
   ypc = np.percentile(y, percentiles, axis=0)
   
   pl.hist2d(x, y, bins=50, normed=True)
   
   for i in xpc:
      pl.axvline(x=i, ls='--', color='w')
   for i in ypc:
      pl.axhline(y=i, ls='--', color='w')
   
   pl.xlabel(xpar)
   pl.ylabel(ypar)
   

def plot_distribution_density(data, xpar, ypar, percentiles=[16, 50, 84]):
   
   x, y = data[xpar][::10], data[ypar][::10]
   
   xpc  = np.percentile(x, percentiles, axis=0)
   ypc = np.percentile(y, percentiles, axis=0)
   
   # Calculate the point density
   xy = np.vstack([x,y])
   z = gaussian_kde(xy)(xy)

   # Sort the points by density, so that the densest points are plotted last
   idx = z.argsort()
   x, y, z = x[idx], y[idx], z[idx]
   
   pl.scatter(x, y, c=z, s=50, edgecolor='')
   

def plot_constraints(constraints, samples, results):
   
   if 'distance' in constraints:
      constraints['d'] = np.array(constraints.pop('distance')) / 44365810.04823812
   
   pars = sorted(constraints.keys())
   
   for i, par in enumerate(pars):
   
      ax = pl.subplot(1, len(pars), i+1)
      
      pc = np.percentile(samples[par], [0.2, 16, 50, 84, 99.8])
      
      #-- plot 1 sigma range as box
      ax.add_patch(
         patches.Rectangle(
            (0.5, pc[1]),
            1.0,
            pc[3]-pc[1],
            fill=False 
         )
      )
         
      #-- plot best fit and 50 percentile fit
      pl.plot([0.5,1.5], [results[par][0], results[par][0]], '--r', lw=1.5)
      pl.plot([0.5,1.5], [results[par][1], results[par][1]], '-b', lw=1.5)
      
      #-- plot 3 sigma range as wiskers
      pl.plot([1.0, 1.0], [pc[0], pc[1]], '-k', lw=1.5, zorder=0)
      pl.plot([1.0, 1.0], [pc[3], pc[4]], '-k', lw=1.5, zorder=0)
      
      #pl.boxplot(samples[par], usermedians=usermedians)
      
      if par in constraints:
         pl.errorbar([1], constraints[par][0], 
                     yerr=[[constraints[par][1]],[constraints[par][2]]],
                     color='r', marker='x', mew=2, lw=2)
      
      ax.axes.get_xaxis().set_visible(False)
      
      if 'teff' in par:
         ticks = ticker.FuncFormatter(lambda x, pos: '{:0.1f}'.format(x/1000.))
         ax.yaxis.set_major_formatter(ticks)
      
      pl.title(par)

 
def plot_fit(obs, obs_err, photbands, pars={}, constraints={}, grids=[], gridnames=[], result='best', plot_components=True, plot_colors=False, plot_constraints=False):
   
   pars = pars.copy()
   
   # use model from 'best' results or 'pc' results
   resi = 0 if result == 'best' else 1

   print(gridnames)
   
   colors = np.array([filters.is_color(p) for p in photbands])
   psystems = np.array([p.split('.')[0] for p in photbands])
   waves = np.array([filters.eff_wave(p) for p in photbands[~colors]])
   
   #-- def system colors
   all_systems = set(psystems)
   colorcycle = cm.rainbow(np.linspace(0, 1, len(all_systems)))
   system_colors = {}
   for p, c in zip(all_systems, colorcycle):
      system_colors[p] = c
   
   abs_xlim = (0.9 * np.min(waves), 1.1 * np.max(waves) )
   
   if not len(list(pars.keys())) == 0:
      ipars = pars.copy()
      for key, value in list(pars.items()):
         ipars[key] = [value[resi]]
         pars[key] = value[resi]
      _ = ipars.pop('d')
      _ = pars.pop('d')
      
      syn, Labs = model.get_itable(grid=grids, photbands=photbands, **ipars)
      syn = syn[:,0]
      
      scale = pars['scale']
   
   # plot the fit of the absolute data
   #====================================
   if not plot_colors and not plot_constraints:
      ax1 = pl.subplot2grid((3,3), (0, 0), colspan=3, rowspan=2)
   else:
      ax1 = pl.subplot2grid((3,3), (0, 0), colspan=2, rowspan=2)
   
   
   if not len(list(pars.keys())) == 0:
      
      #-- synthetic integrated fluxes
      pl.plot(waves, scale*syn[~colors], 'xr')
      
      #-- if possible, plot a non integrated model
      #   can only be done if provided gridnames are not paths to integrated files.
      if len(gridnames) > 0 and not os.path.isfile(gridnames[0]):
         #-- synthetic model
         ebv = pars['ebv']
         wave, flux = model.get_table(grid=gridnames, **pars)
         flux = reddening.redden(flux, wave=wave, ebv=ebv, rtype='flux', law='fitzpatrick2004')
         
         #ascii.write([wave, scale*flux], 'binary_model.txt', names=['wave', 'flux'], overwrite=True)
         
         pl.plot(wave, scale*flux, '-r')
         
         #-- plot components
         if plot_components and 'teff2' in pars:
            wave, flux = model.get_table(grid=gridnames[0], teff=pars['teff'], logg=pars['logg'], rad=pars['rad'], ebv=pars['ebv'])
            flux = reddening.redden(flux, wave=wave, ebv=ebv, rtype='flux', law='fitzpatrick2004')
            #ascii.write([wave, scale*flux], 'primary_model.txt', names=['wave', 'flux'], overwrite=True)
            pl.plot(wave, scale*flux, '--g')
            
            wave, flux = model.get_table(grid=gridnames[1], teff=pars['teff2'], logg=pars['logg2'], rad=pars['rad2'], ebv=pars['ebv'])
            flux = reddening.redden(flux, wave=wave, ebv=ebv, rtype='flux', law='fitzpatrick2004')
            #ascii.write([wave, scale*flux], 'secondary_model.txt', names=['wave', 'flux'], overwrite=True)
            pl.plot(wave, scale*flux, '--b')
         
         
      
   wave, flux, error, band = [], [], [], []
   for system in all_systems:
      s = np.where((psystems == system) & (~colors))
      if len(obs[s]) == 0: continue
   
      w = np.array([filters.eff_wave(p) for p in photbands[s]])
      pl.errorbar(w, obs[s], yerr=obs_err[s], ls='', marker='o', 
                  color=system_colors[system], label=system)
      
      wave.extend(w)
      flux.extend(obs[s])
      error.extend(obs_err[s])
      band.extend([system for i in w])
      
    
   ascii.write([wave, flux, error, band], 'observations.txt', names=['wave', 'flux', 'error', 'band'], overwrite=True)
   
      
   pl.xlim(abs_xlim)
   pl.ylim([0.9*np.min(obs[~colors]), 1.1*np.max(obs[~colors])])
   pl.legend(loc='best', prop={'size': 9}, numpoints=1)
   ax1.set_xscale("log", nonposx='clip')
   ax1.set_yscale("log", nonposy='clip')
   
   pl.ylabel('Absolute Flux')
   
   #-- add band names to top of axis
   ax2 = ax1.twiny()
   ax2.set_xscale("log", nonposx='clip')
   ax2.set_xlim(abs_xlim)
   waves = np.array([filters.eff_wave(p) for p in photbands[~colors]])
   bandnames = np.array([b.split('.')[-1] for b in photbands[~colors]])
   ax2.set_xticks(waves)
   ax2.set_xticklabels(bandnames)
   ax2.xaxis.tick_top()
   pl.minorticks_off()
   
   
   
   # plot O-C of the absolute measurements
   #====================================
   if not plot_colors and not plot_constraints:
      ax = pl.subplot2grid((3,3), (2, 0), colspan=3, rowspan=1)
   else:
      ax = pl.subplot2grid((3,3), (2, 0), colspan=2, rowspan=1)
   
   for system in all_systems:
      s = np.where((psystems == system) & (~colors))
      if len(obs[s]) == 0: continue
      
      w = np.array([filters.eff_wave(p) for p in photbands[s]])
      y = ((obs - syn*scale) / obs)[s]
      yerr = (obs_err / obs)[s]
      
      y = 2.5*np.log10(syn[s]*scale / obs[s])
      yerr = 2.5 / np.log(10.0) * obs_err[s] / obs[s]
      
      pl.errorbar(w, -y , yerr=yerr, ls='', marker='o', color=system_colors[system])
   
   pl.figtext(0.96, 0.96, '$\chi^2$ = {:0.2f}'.format(pars['chi2']), ha='right')
   
   pl.axhline(y=0, color='k', ls='--')
   
   pl.xlim(abs_xlim)
   pl.gca().set_xscale('log',nonposx='clip')
   
   pl.ylabel('(O-C) (mag)')
   pl.xlabel('Wavelength (AA)')
   

   # plot O-C of the color fits
   #====================================
   if len(obs[colors]) > 0 and plot_colors:
      
      ax = pl.subplot2grid((3,3), (0, 2), colspan=1, rowspan=2)
      
      y = np.array(list(range(len(obs[colors]))))
      x = obs[colors] - syn[colors]
      x_err = obs_err[colors]
      oc_sys = psystems[colors]
      
      for system in all_systems:
         s = np.where(oc_sys == system)
         if len(x[s]) == 0: continue
      
         pl.errorbar(x[s], y[s], xerr=x_err[s], ls='', marker='o',
                     color=system_colors[system])
      
      pl.axvline(x=0, color='k', ls='--')
      
      yticknames = [b.split('.')[1] for b in photbands[colors]]
      pl.yticks(y, yticknames)
      pl.ylim([-y[-1]*0.1, y[-1]*1.1])
      
      ax.yaxis.tick_right()
      ax.xaxis.tick_top()
      ax.xaxis.set_label_position('top') 
      
      pl.xlabel('O-C')
   
   
   # plot fractional O-C of the constraints
   #=======================================
   if plot_constraints:
      ax = pl.subplot2grid((3,3), (2, 2), colspan=1, rowspan=1)
      
      x, xticknames = [], []
      
      if 'q' in constraints:
         
         q, ql, qu = constraints['q']
         
         ax.plot([1], pars['q'], 'k_', ms=15, mew=2)
         ax.errorbar([1], [q], yerr = [[ql], [qu]], marker='o')
         
         x.append(1)
         xticknames.append('q')
      
      
      if 'distance' in constraints:
         
         ax2 = ax.twinx()
         
         d  = constraints['distance'][0] / 44365810.04823812
         de = constraints['distance'][1] / 44365810.04823812
         
         
         ax2.plot([2], pars['d'], 'k_', ms=15, mew=2)
         ax2.errorbar([2], [d], yerr = de, marker='o')
         
         x.append(2)
         xticknames.append('distance')
      
      pl.xlim([0.5, 2.5])
      pl.xticks(x, xticknames)
   
   
   pl.figtext(0.03, 0.96, 'model = {}'.format(result))



default_single = """
# photometry file with index to the columns containing the photbands, observations and errors
objectname: <objectname>
photometryfile: <photfilename>
photband_index: band
obs_index: flux
err_index: eflux
photband_exclude: []
# parameters to fit and the limits on them in same order as parameters
pnames: [teff, logg, rad, ebv]
limits:
- [20000, 60000]
- [5.80, 5.80]
- [0.01, 0.5]
- [0, 0.10]
# constraints on distance and mass ratio is known
constraints: 
  parallax: [<plx>, <e_plx>]
# added constraints on derived properties as mass, luminosity, luminosity ratio
derived_limits: {}
# path to the model grids with integrated photometry
grids: 
- tmap
# setup for the MCMC algorithm
nwalkers: 100    # total number of walkers
nsteps: 1000     # steps taken by each walker (not including burn-in)
nrelax: 250      # burn-in steps taken by each walker
a: 10            # relative size of the steps taken
# set the percentiles for the error determination 
percentiles: [0.2, 50, 99.8] # 16 - 84 corresponds to 1 sigma
# output options
resultfile: <objectname>_results_single.csv   # filepath to write results
plot1:
 type: sed_fit
 result: pc
 path: <objectname>_sed_single.png
plot2:
 type: distribution
 show_best: true
 path: <objectname>_distribution_single.png
 parameters: ['teff', 'rad', 'L', 'ebv', 'd', 'mass']
"""

default_double = """
# photometry file with index to the columns containing the photbands, observations and errors
objectname: <objectname>
photometryfile: <photfilename>
photband_index: band
obs_index: flux
err_index: eflux
photband_exclude: []
# parameters to fit and the limits on them in same order as parameters
pnames: [teff, logg, rad, teff2, logg2, rad2, ebv]
limits:
- [3500, 10000]
- [4.31, 4.31]
- [0.01, 2.5]
- [20000, 50000]
- [5.8, 5.8]
- [0.01, 0.5]
- [0, 0.10]
# constraints on distance and mass ratio if known
constraints:
  parallax: [<plx>, <e_plx>]
# added constraints on derived properties as mass, luminosity, luminosity ratio
derived_limits: {}
# path to the model grids with integrated photometry
grids: 
- kurucz2
- tmap
# setup for the MCMC algorithm
nwalkers: 100    # total number of walkers
nsteps: 500     # steps taken by each walker (not including burn-in)
nrelax: 250      # burn-in steps taken by each walker
a: 10            # relative size of the steps taken
# set the percentiles for the error determination 
percentiles: [16, 50, 84] # 16 - 84 corresponds to 1 sigma
# output options
resultfile: <objectname>_results_binary.csv   # filepath to write results
plot1:
 type: sed_fit
 path: <objectname>_sed_binary.png
plot2:
 type: distribution
 path: <objectname>_distribution_primary.png
 parameters: ['teff', 'rad', 'teff2', 'rad2', 'ebv', 'd']
#plot3:
# type: distribution
# #path: <objectname>_distribution_derived.png
# parameters: ['mass', 'L', 'mass2', 'L2', 'q']
"""



import os

import numpy as np
 
from astropy.io import ascii
basedir = os.path.dirname(__file__)

def is_color(photband):
   """
   Return true if the photometric passband is actually a color.
   
   @param photband: name of the photometric passband
   @type photband: string
   @return: True or False
   @rtype: bool
   """
   if '-' in photband.split('.')[1]:
      return True
   elif photband.split('.')[1].upper() in ['M1','C1']:
      return True
   else:
      return False


def eff_wave(photband):
   """
   Returns the effective wavelength of the pass band in angstrom
   
   @param photband: name of the photometric passband
   @type photband: string
   @return: effective wavelength
   @rtype: float
   """
   
   data = ascii.read(os.path.join(basedir, 'zeropoints.dat'), comment="\s*#")
   
   s = np.where(data['photband'] == photband)
   
   return data['eff_wave'][s][0]


def mag2flux(mag, error, photband):
   """
   Converts a magnitude in a given photband to a flux
   
   Flux is returned in units of erg/s/cm2/AA
   """

   # Todo: Deal with stromgren colours!!!

   data = ascii.read(os.path.join(basedir, 'zeropoints.dat'), comment="\s*#")
   
   if not isinstance(photband, str):
      flux, err = np.zeros_like(mag), np.zeros_like(mag)
      for i, (m, e, b) in enumerate(zip(mag, error, photband)):
         
         s = np.where(data['photband'] == b)

         F0 = data['Flam0'][s][0]
         zpcor = data['zp_corr'][s][0]
         
         f = 10**(-(m-zpcor)/2.5)*F0
         fe = np.log(10) * e / 2.5 * f
         
         flux[i] = f
         err[i] = fe
   
   else:
      s = np.where(data['photband'] == photband)
      data = data[s]

      F0 = data['Flam0'][0]
      zpcor = data['zp_corr'][0]
      
      flux = 10**(-(mag-zpcor)/2.5)*F0
      err = np.log(10) * error / 2.5 * flux

      if hasattr(err, '__iter__'):
         err = np.where(err == 0, flux / 10, err)
      elif err == 0:
         err = flux / 10
   
   return flux, err


def flux2mag(flux, error, photband):
   """
   Converts a flux in a given photband to a magnitude
   
   Flux has to be provided in units of erg/s/cm2/AA
   """
   data = ascii.read(os.path.join(basedir, 'zeropoints.dat'), comment="\s*#")

   if not isinstance(photband, str):
      mag, err = np.zeros_like(flux), np.zeros_like(flux)
      for i, (f, e, b) in enumerate(zip(flux, error, photband)):
         
         s = np.where(data['photband'] == b)
         
         F0 = data['Flam0'][s][0]
         zpcor = data['zp_corr'][s][0]
         
         m = -2.5*np.log10(f/F0)+zpcor
         me = 2.5 / np.log(10) * e / f
         
         mag[i] = m
         err[i] = me
   
   else:
      s = np.where(data['photband'] == photband)
      data = data[s]
      
      F0 = data['Flam0'][0]
      zpcor = data['zp_corr'][0]
      
      mag = -2.5*np.log10(flux/F0)+zpcor
      err = 2.5 / np.log(10) * error / flux
   
   return  mag, err


if __name__=="__main__":
   
   print((mag2flux(13.0, 0.01, '2MASS.H')))



import numpy as np
from scipy import ndimage

def create_pixeltypegrid(grid_pars,grid_data):
   """
   Creates pixelgrid and arrays of axis values.
   
   Starting from:
      * grid_pars: 2D numpy array, 1 column per parameter, unlimited number of cols
      * grid_data: 2D numpy array, 1 column per variable, data corresponding to the rows in grid_pars
   
   example: interpolation in a 3D grid containing stellar evolution models. Say we have as
   input parameters mass, age and metalicity, and want to obtain teff and logg as variables.
   
   grid_pars =
      +------+-----+------+
      | mass | age | Fe/H |
      +------+-----+------+
      | 1.0  | 1.0 | -0.5 |
      +------+-----+------+
      | 2.0  | 1.0 | -0.5 |
      +------+-----+------+
      | 1.0  | 2.0 | -0.5 |
      +------+-----+------+
      | 2.0  | 2.0 | -0.5 |
      +------+-----+------+
      | 1.0  | 1.0 |  0.0 |
      +------+-----+------+
      | 2.0  | 1.0 |  0.0 |
      +------+-----+------+
      |...   |...  |...   |
      +------+-----+------+
      
   grid_data = 
      +------+------+
      | teff | logg |
      +------+------+
      | 5000 | 4.45 |
      +------+------+
      | 6000 | 4.48 |
      +------+------+
      |...   |...   |
      +------+------+
   
   The resulting grid will be rectangular and complete. This means that every
   combination of unique values in grid_pars should exist. If this is not the
   case, a +inf value will be inserted in grid_data at all locations that are 
   missing!

   
   :param grid_pars: Npar x Ngrid array of parameters
   :type grid_pars: array
   :param grid_data: Ndata x Ngrid array of data
   :type grid_data: array
   
   :return: axis values and pixelgrid
   :rtype: array, array
   """

   uniques = [np.unique(column, return_inverse=True) for column in grid_pars]
   #[0] are the unique values, [1] the indices for these to recreate the original array

   axis_values = [uniques_[0] for uniques_ in uniques]
   unique_val_indices = [uniques_[1] for uniques_ in uniques]
   
   data_dim = np.shape(grid_data)[0]

   par_dims   = [len(uv[0]) for uv in uniques]

   par_dims.append(data_dim)
   pixelgrid = np.ones(par_dims)
   
   # We put np.inf as default value. If we get an inf, that means we tried to access
   # a region of the pixelgrid that is not populated by the data table
   pixelgrid[pixelgrid==1] = np.inf
   
   # now populate the multiDgrid
   indices = [uv[1] for uv in uniques]
   pixelgrid[indices] = grid_data.T
   
   return axis_values, pixelgrid

def interpolate(p, axis_values, pixelgrid):
   """
   Interpolates in a grid prepared by create_pixeltypegrid().
   
   p is an array of parameter arrays
   each collumn contains the value for the corresponding parameter in grid_pars
   each row contains a set of model parameters for wich the interpolated values
   in grid_data are requested.
   
   example: continue with stellar evolution models used in create_pixeltypegrid
   
   p = 
      +------+-----+-------+
      | mass | age | Fe/H  | 
      +------+-----+-------+
      | 1.21 | 1.3 | 0.24  |
      +------+-----+-------+
      | 1.57 | 2.4 | -0.15 |
      +------+-----+-------+
      |...   |...  |...    |
      +------+-----+-------+
      
   >>> p = np.array([[1.21, 1.3, 0.24], [1.57, 2.4, -0.15]])
   >>> interpolate(p, axis_values, pixelgrid)
   >>> some output
   
   :param p: Npar x Ninterpolate array containing the points which to
             interpolate in axis_values
   :type p: array
   :param axis_values: output from create_pixeltypegrid
   :type axis_values: array
   :param pixelgrid: output from create_pixeltypegrid
   :type pixelgrid: array
   
   :return: Ndata x Ninterpolate array containing the interpolated values
            in pixelgrid
   :rtype: array
   
   """
   # convert requested parameter combination into a coordinate
   #p_ = [np.searchsorted(av_,val) for av_, val in zip(axis_values,p)]
   # we force the values to be inside the grid, to avoid edge-effect rounding
   # (e.g. 3.099999 is edge, while actually it is 3.1). For values below the
   # lowest value, this is automatically done via searchsorted (it return 0)
   # for values higher up, we need to force it
   
   p_ = []
   for av_,val in zip(axis_values,p):
      indices = np.searchsorted(av_,val)
      indices[indices==len(av_)] = len(av_)-1
      p_.append(indices)
   

   #-- The type of p is changes to the same type as in axis_values to catch possible rounding errors
   #   when comparing float64 to float32.
   for i, ax in enumerate(axis_values):
      p[i] = np.array(p[i], dtype = ax.dtype)
   
   #-- Convert requested parameter combination into a coordinate
   p_ = np.array([np.searchsorted(av_,val) for av_, val in zip(axis_values,p)])
   lowervals_stepsize = np.array([[av_[p__-1], av_[p__]-av_[p__-1]] \
                           for av_, p__ in zip(axis_values,p_)])
   p_coord = (p-lowervals_stepsize[:,0])/lowervals_stepsize[:,1] + np.array(p_)-1

   # interpolate
   return np.array([ndimage.map_coordinates(pixelgrid[...,i],p_coord, order=1, prefilter=False) \
               for i in range(np.shape(pixelgrid)[-1])])




import pytest

import os

import numpy as np

from speedyfit import fileio

default = """
# photometry file and column indices
photometryfile: path/to/file.dat
photband_index: 0
obs_index: 1
err_index: 2
# parameters to fit and the limits on them in same order as parameters
pnames: [teff, logg, rad, teff2, logg2, rad2, ebv]
limits:
- [3500, 6000]
- [3.5, 5.0]
- [0.7, 1.5]
- [20000, 40000]
- [4.5, 6.5]
- [0.05, 0.3]
- [0, 0.02]
# constraints on distance and mass ratio if known
constraints:
  q: [3.03, 0.2]
  distance: [600, 50] # in parsec
# constraints on derived properties
derived_limits:
  mass: [0.5, 1.0]
  mass2: [0.1, 1.0]
# path to the model grids with integrated photometry
grids:
- path/to/grid/1.fits
- path/to/grid/2.fits
# setup for the MCMC algorithm
nwalkers: 100    # total number of walkers
nsteps: 2000     # steps taken by each walker (not including burn-in)
nrelax: 500      # burn-in steps taken by each walker
a: 10            # relative size of the steps taken
# output options
datafile: none   # filepath to write results of all walkers
"""


class TestFileIO:

    def test_write2fits(self):

        data = np.array([(1, 2), (2, 3)], dtype=[('a', 'f8'), ('b', 'f8')])

        try:
            fileio.write2fits(data, 'testfile.fits', setup=default)

            samples, setup = fileio.read_fits('testfile.fits')

            assert setup == default

            for name in data.dtype.names:
                for v1, v2 in zip(data[name], samples[name]):
                    assert v1 == v2

            os.remove('testfile.fits')

        finally:
            if os.path.isfile('testfile.fits'):
                os.remove('testfile.fits')



import pytest

import numpy as np

from speedyfit import statfunc


class TestMCMC:

    def test_derived_properties(self):
        theta = (26000, 5.8, 0.1429, 5771, 4.438, 1, 0.02)
        pnames = ('teff', 'logg', 'rad', 'teff2', 'logg2', 'rad2', 'ebv')

        derived_properties = statfunc.get_derived_properties_binary(theta, pnames)

        assert derived_properties['mass'] == pytest.approx(0.47, 0.01), \
            "sdB mass wrongly calculated: {}".format(derived_properties['mass'])

        assert derived_properties['mass2'] == pytest.approx(1.00, 0.01), \
            "Solar mass wrongly calculated: {}".format(derived_properties['mass2'])

        assert derived_properties['q'] == pytest.approx(0.47, 0.001), \
            "q wrongly calculated: {}".format(derived_properties['q'])

    def test_stat_chi2(self):
        meas = np.array([1.0, 2.0, 3.0])
        e_meas = np.array([0.25, 0.30, 0.15])
        syn = np.array([1.5, 1.89, 3.6])

        # -- colors (none scaled)
        colors = np.array([True, True, True])
        chi2, scale, e_scale = statfunc.stat_chi2(meas, e_meas, colors, syn)

        assert chi2 == pytest.approx(20.1344, 0.0001), \
            "Chi2 for colours not correct ({} != {})".format(chi2, 20.1344)

        # -- absolute magnitudes (scaled)
        colors = np.array([False, False, False])
        chi2, scale, e_scale = statfunc.stat_chi2(meas, e_meas, colors, syn)

        assert chi2 == pytest.approx(3.32834, 0.0001), \
            "Chi2 for colours not correct ({} != {})".format(chi2, 3.3283)

        # -- added distance
        constraints = {'distance': (1.5, 0.15)}
        colors = np.array([False, False, False])
        chi2, scale, e_scale = statfunc.stat_chi2(meas, e_meas, colors, syn, **constraints)

        assert chi2 == pytest.approx(25.2343, 0.0001), "Chi2 for distance not correct ({} != {})".format(chi2, 25.2343)



import pytest

import numpy as np

from speedyfit import mcmc, statfunc


class TestMCMC:

    # def setUp(self):

    # self.evolution_model = 'mist'
    # self.variables = ['log_L', 'log_Teff', 'log_g', 'M_H']
    # self.limits = [(0.2, 1.1), (-1.0, 0.25), (5.0, 9.0)]
    # self.obs = np.array([-0.55, 3.67, 4.50, -0.35])
    # self.obs_err = np.array([0.15, 0.05, 0.50, 0.40])

    def test_prior(self):
        theta = (26000, 5.8, 0.1429, 5771, 4.438, 1, 0.02)
        pnames = ('teff', 'logg', 'rad', 'teff2', 'logg2', 'rad2', 'ebv')

        limits = np.array([(20000, 40000), (5.0, 6.5), (0.05, 0.25),
                           (4000, 8000), (4.0, 5.0), (0.50, 2.10),
                           (0.00, 0.05)])

        derived_limits = {'q': (0.1, 1.0), 'mass': (0.40, 0.55), 'mass2': (0.70, 2.0)}

        derived_properties = statfunc.get_derived_properties_binary(theta, pnames)

        prior = mcmc.lnprior(theta, derived_properties, limits,
                             pnames=pnames, derived_limits=derived_limits)

        assert prior == 0, "theta within limits, expected priod = 0, was {}".format(prior)

        theta = (15000, 5.8, 0.1429, 5771, 4.438, 1, 0.02)
        derived_properties = statfunc.get_derived_properties_binary(theta, pnames)

        prior = mcmc.lnprior(theta, derived_properties, limits,
                             pnames=pnames, derived_limits=derived_limits)

        assert prior == -np.inf, "theta out of limits, expected prior = -inf, was {}".format(prior)

        theta = (26000, 5.8, 0.10, 5771, 4.438, 1, 0.02)
        derived_properties = statfunc.get_derived_properties_binary(theta, pnames)

        prior = mcmc.lnprior(theta, derived_properties, limits,
                             pnames=pnames, derived_limits=derived_limits)

        assert prior == -np.inf, "mass out of limits, expected prior = -inf, was {}".format(prior)

    # def test_likelihood(self):

    # photbands = ['STROMGREN.U', 'STROMGREN.B']
    # gridfilename = '/home/joris/Python/ivsdata/sedtables/modelgrids/ikurucz93_z0.0_k2odfnew_sed_lawfitzpatrick2004_Rv3.10.fits'

    # axis_values, grid_pars, pixelgrid, grid_names = model.prepare_grid(photbands, gridfilename,
    # teffrange=(4000, 5000),loggrange=(3.5, 4.0),
    # ebvrange=(0.0, 0.02),
    # variables=['teff','logg','ebv'])

    # grids = [[axis_values, pixelgrid],]

    # theta = (4700, 3.8, 0.01)
    # pnames = ['teff', 'logg', 'ebv']
    # derived_properties = statfunc.get_derived_properties(theta, pnames)

    # y = np.array([5.0, 6.0])
    # yerr = np.array([0.5, 0.75])

    # kwargs = dict(pnames = pnames,
    # photbands = photbands,
    # grid=grids)

    # print mcmc.lnlike(theta, derived_properties, y, yerr, **kwargs)

    # self.assertTrue(True)



# -*- coding: utf-8 -*-
import numpy as np

import warnings


def get_derived_properties(**pars):
    """
    Function that will derive several properties based on the chosen models
   
    Currently the following properties are calculated:
    - mass
    - mass1
    - mass2
    - q
   
    returns dictionary of all properties that could be calculated.
    """

    derived_properties = {}

    GG = 6.67384e-08
    Rsol = 69550800000.0
    Msol = 1.988547e+33

    # -- derive masses
    if 'rad' in pars and 'logg' in pars:
        mass = 10 ** pars['logg'] * (pars['rad'] * Rsol) ** 2 / GG
        derived_properties['mass'] = mass / Msol

    if 'rad' in pars and 'g' in pars:
        mass = pars['g'] * (pars['rad'] * Rsol) ** 2 / GG
        derived_properties['mass'] = mass / Msol

    if 'rad2' in pars and 'logg2' in pars:
        mass = 10 ** pars['logg2'] * (pars['rad2'] * Rsol) ** 2 / GG
        derived_properties['mass2'] = mass / Msol

    if 'rad2' in pars and 'g2' in pars:
        mass = pars['g2'] * (pars['rad2'] * Rsol) ** 2 / GG
        derived_properties['mass2'] = mass / Msol

        # -- derive radii
    if 'mass' in pars and 'logg' in pars:
        rad = np.sqrt(GG * pars['mass'] * Msol / 10 ** pars['logg'])
        derived_properties['rad'] = rad / Rsol

    if 'mass' in pars and 'g' in pars:
        rad = np.sqrt(GG * pars['mass'] * Msol / pars['g'])
        derived_properties['rad'] = rad / Rsol

    if 'mass2' in pars and 'logg2' in pars:
        rad = np.sqrt(GG * pars['mass2'] * Msol / 10 ** pars['logg2'])
        derived_properties['rad2'] = rad / Rsol

    if 'mass2' in pars and 'g2' in pars:
        rad = np.sqrt(GG * pars['mass2'] * Msol / pars['g2'])
        derived_properties['rad2'] = rad / Rsol

    # -- derive mass ratio
    if 'mass' in derived_properties and 'mass2' in derived_properties:
        derived_properties['q'] = derived_properties['mass'] / derived_properties['mass2']

    if 'mass' in pars and 'mass2' in pars:
        derived_properties['q'] = pars['mass'] / pars['mass2']

    # -- derive luminosity ratio
    if 'rad' in pars and 'teff' in pars and 'rad2' in pars and 'teff2' in pars:
        l1 = (pars['rad'] ** 2 * pars['teff'] ** 4)
        l2 = (pars['rad2'] ** 2 * pars['teff2'] ** 4)
        derived_properties['lr'] = l1 / l2

    if 'mass' in pars and 'teff' in pars and 'mass2' in pars and 'teff2' in pars:
        l1 = (derived_properties['rad'] ** 2 * pars['teff'] ** 4)
        l2 = (derived_properties['rad2'] ** 2 * pars['teff2'] ** 4)
        derived_properties['lr'] = l1 / l2

    # -- derive radius ratio
    if 'rad' in pars and 'rad2' in pars:
        derived_properties['rr'] = pars['rad'] / pars['rad2']

    if 'rad' in derived_properties and 'rad2' in derived_properties:
        derived_properties['rr'] = derived_properties['rad'] / derived_properties['rad2']

    # -- add empty values for luminosity and distance to prevent problems with
    #   failed models
    derived_properties.update({'d': 0, 'L': 0})
    if 'rad2' in pars:
        derived_properties['L2'] = 0
    if hasattr(pars['teff'], '__iter__'):
        derived_properties['d'] = np.zeros_like(pars['teff'])
        derived_properties['L'] = np.zeros_like(pars['teff'])
        if 'rad2' in pars:
            derived_properties['L2'] = np.zeros_like(pars['teff'])

    return derived_properties


# def get_derived_properties(theta, pnames):
# """
# Function that will derive several properties based on the chosen models

# Currently the following properties are calculated:
# - mass
# - mass1
# - mass2
# - q

# returns dictionary of all properties that could be calculated.
# """

# derived_properties = {}

# GG = 6.67384e-08
# Rsol = 69550800000.0
# Msol = 1.988547e+33

##-- derive masses
# if 'rad' in pnames and 'logg' in pnames:
# mass = 10**theta[pnames.index('logg')] * (theta[pnames.index('rad')] * Rsol)**2 / GG
# derived_properties['mass'] = mass / Msol

# if 'rad' in pnames and 'g' in pnames:
# mass = theta[pnames.index('g')] * (theta[pnames.index('rad')] * Rsol)**2 / GG
# derived_properties['mass'] = mass / Msol

# if 'rad2' in pnames and 'logg2' in pnames:
# mass = 10**theta[pnames.index('logg2')] * (theta[pnames.index('rad2')] * Rsol)**2 / GG
# derived_properties['mass2'] = mass / Msol

# if 'rad2' in pnames and 'g2' in pnames:
# mass = theta[pnames.index('g2')] * (theta[pnames.index('rad2')] * Rsol)**2 / GG
# derived_properties['mass2'] = mass / Msol

##-- derive radii
# if 'mass' in pnames and 'logg' in pnames:
# rad = np.sqrt(GG * theta[pnames.index('mass')] * Msol / 10**theta[pnames.index('logg')])
# derived_properties['rad'] = rad / Rsol

# if 'mass' in pnames and 'g' in pnames:
# rad = np.sqrt(GG * theta[pnames.index('mass')] * Msol / theta[pnames.index('g')])
# derived_properties['rad'] = rad / Rsol

# if 'mass2' in pnames and 'logg2' in pnames:
# rad = np.sqrt(GG * theta[pnames.index('mass2')] * Msol / 10**theta[pnames.index('logg2')])
# derived_properties['rad2'] = rad / Rsol

# if 'mass2' in pnames and 'g2' in pnames:
# rad = np.sqrt(GG * theta[pnames.index('mass2')] * Msol / theta[pnames.index('g2')])
# derived_properties['rad2'] = rad / Rsol

##-- derive mass ratio
# if 'mass' in derived_properties and 'mass2' in derived_properties:
# derived_properties['q'] = derived_properties['mass'] / derived_properties['mass2']

# if 'mass' in pnames and 'mass2' in pnames:
# derived_properties['q'] = theta[pnames.index('mass')] / theta[pnames.index('mass2')]

##-- derive luminosity ratio
# if 'rad' in pnames and 'teff' in pnames and 'rad2' in pnames and 'teff2' in pnames:
# l1 = ( theta[pnames.index('rad')]**2 * theta[pnames.index('teff')]**4 )
# l2 = ( theta[pnames.index('rad2')]**2 * theta[pnames.index('teff2')]**4 )
# derived_properties['lr'] = l1 / l2

# if 'mass' in pnames and 'teff' in pnames and 'mass2' in pnames and 'teff2' in pnames:
# l1 = ( derived_properties['rad']**2 * theta[pnames.index('teff')]**4 )
# l2 = ( derived_properties['rad2']**2 * theta[pnames.index('teff2')]**4 )
# derived_properties['lr'] = l1 / l2

##-- add empty values for luminosity and distance to prevent problems with
##   failed models
# derived_properties.update({'d':0, 'L':0})
# if 'rad2' in pnames:
# derived_properties['L2'] = 0

# return derived_properties

def get_derived_properties_binary(theta, pnames):
    """
    Function that will derive several properties based on the chosen models
   
    Currently the following properties are calculated:
    - mass
    - mass1
    - mass2
    - q
   
    returns dictionary of all properties that could be calculated.
    """

    derived_properties = {}

    GG = 6.67384e-08
    Rsol = 69550800000.0
    Msol = 1.988547e+33

    # -- derive masses
    if 'g' in pnames:
        m1 = theta[pnames.index('g')] * (theta[pnames.index('rad')] * Rsol) ** 2 / GG
    else:
        m1 = 10 ** theta[pnames.index('logg')] * (theta[pnames.index('rad')] * Rsol) ** 2 / GG
    derived_properties['mass'] = m1 / Msol

    if 'g2' in pnames:
        m2 = theta[pnames.index('g2')] * (theta[pnames.index('rad2')] * Rsol) ** 2 / GG
    else:
        m2 = 10 ** theta[pnames.index('logg2')] * (theta[pnames.index('rad2')] * Rsol) ** 2 / GG
    derived_properties['mass2'] = m2 / Msol

    # -- derive mass ratio
    derived_properties['q'] = m1 / m2

    # -- derive luminosity ratio
    l1 = (theta[pnames.index('rad')] ** 2 * theta[pnames.index('teff')] ** 4)
    l2 = (theta[pnames.index('rad2')] ** 2 * theta[pnames.index('teff2')] ** 4)
    derived_properties['lr'] = l1 / l2

    # -- add empty values for luminosity and distance to prevent problems with
    #   failed models
    derived_properties.update({'d': 0, 'L': 0, 'L2': 0})

    return derived_properties


def get_derived_properties_single(theta, pnames):
    """
    Function that will derive several properties based on the chosen models
   
    Currently the following properties are calculated:
    - mass
    - mass1
    - mass2
    - q
   
    returns dictionary of all properties that could be calculated.
    """

    derived_properties = {}

    GG = 6.67384e-08
    Rsol = 69550800000.0
    Msol = 1.988547e+33

    # -- derive masses
    if 'g' in pnames:
        mass = theta[pnames.index('g')] * (theta[pnames.index('rad')] * Rsol) ** 2 / GG
    else:
        mass = 10 ** theta[pnames.index('logg')] * (theta[pnames.index('rad')] * Rsol) ** 2 / GG
    derived_properties['mass'] = mass / Msol

    # -- add empty values for luminosity and distance to prevent problems with
    #   failed models
    derived_properties.update({'d': 0, 'L': 0})

    return derived_properties


def stat_chi2(meas, e_meas, colors, syn, pars, **kwargs):
    """
    Calculate Chi2 and compute angular diameter.
   
    Colors and absolute fluxes are used to compute the Chi2, only absolute
    fluxes are used to compute angular diameter. If no absolute fluxes are
    given, the angular diameter is set to 0.

    If constraints are given in the kwargs they are included in the Chi2
    calculation. Accepted constrains are distance and q (mass ratio). Both
    should be given as a (value, error) tuple.

    @param meas: array of measurements
    @type meas: 1D array
    @param e_meas: array containing measurements errors
    @type e_meas: 1D array
    @param colors: boolean array separating colors (True) from absolute fluxes (False)
    @type colors: 1D boolean array
    @param syn: synthetic fluxes and colors
    @type syn: 1D array
    @param full_output: set to True if you want individual chisq
    @type full_output: boolean
    @return: chi-square, scale, e_scale
    @rtype: float,float,float
    """

    # check for NaNs in the observations
    nani = np.isnan(meas) | np.isnan(e_meas)
    if any(nani):
        warnings.warn('{} of the observed magnitudes are NaN values! (NaN values are ignored)'.format(np.sum(nani)))
    meas, e_meas, colors, syn = meas[~nani], e_meas[~nani], colors[~nani], syn[~nani]

    # First deal with Chi2 of the observations
    # =========================================
    # -- if syn represents only one measurement
    if sum(~colors) > 0:
        ratio = (meas / syn)[~colors]
        weights = (meas / e_meas)[~colors]
        # -- weighted average and standard deviation
        scale = np.average(ratio, weights=weights)
        e_scale = np.sqrt(np.dot(weights, (ratio - scale) ** 2) / weights.sum())
    else:
        scale, e_scale = 0, 0
    # -- we don't need to scale the colors, only the absolute fluxes
    chisq = np.where(colors, (syn - meas) ** 2 / e_meas ** 2, (syn * scale - meas) ** 2 / e_meas ** 2)

    # Then add Chi2 of derived properties as distance, mass ratio, ...
    # =================================================================
    derived_properties = kwargs.get('derived_properties', {})
    constraints = kwargs.get('constraints', {})

    for con, (c, c_m, c_p) in list(constraints.items()):
        syn_c = None
        if con == 'distance' and sum(~colors) > 0:
            # distance can only be constrained if there is at least 1 absolute measurement
            c, c_m, c_p = 1. / c ** 2, 2. * c_m / c ** 3, 2. * c_p / c ** 3
            syn_c = scale

        elif con in pars:
            # constrained on a fitted parameter
            syn_c = pars[con]

        elif con in derived_properties:
            # constrained on a derived parameter
            syn_c = derived_properties[con]

        if not syn_c is None:
            chi2_c = np.where(syn_c < c, (syn_c - c) ** 2 / c_m ** 2, (syn_c - c) ** 2 / c_p ** 2)

            # append to chisq array
            chisq = np.append(chisq, chi2_c)

    return chisq.sum(axis=0), scale, e_scale


def stat_chi2_multiple(meas, e_meas, colors, syn, pars, **kwargs):
    """
    Calculate Chi2 and compute angular diameter.

    Colors and absolute fluxes are used to compute the Chi2, only absolute
    fluxes are used to compute angular diameter. If no absolute fluxes are
    given, the angular diameter is set to 0.

    If constraints are given in the kwargs they are included in the Chi2
    calculation. Accepted constrains are distance and q (mass ratio). Both
    should be given as a (value, error) tuple.

    @param meas: array of measurements
    @type meas: 1D array
    @param e_meas: array containing measurements errors
    @type e_meas: 1D array
    @param colors: boolean array separating colors (True) from absolute fluxes (False)
    @type colors: 1D boolean array
    @param syn: synthetic fluxes and colors
    @type syn: 1D array
    @param full_output: set to True if you want individual chisq
    @type full_output: boolean
    @return: chi-square, scale, e_scale
    @rtype: float,float,float
    """

    # First deal with Chi2 of the observations
    # =========================================
    # -- if syn represents a grid of measurements
    if sum(~colors) > 0:
        ratio = (meas / syn)[~colors]
        weights = (meas / e_meas)[~colors]
        # -- weighted average and standard deviation
        scale = np.average(ratio, weights=weights.reshape(-1), axis=0)
        e_scale = np.sqrt(np.dot(weights.T, (ratio - scale) ** 2) / weights.sum(axis=0))[0]
    else:
        scale, e_scale = np.zeros(syn.shape[1]), np.zeros(syn.shape[1])

    # -- we don't need to scale the colors, only the absolute fluxes
    chisq = np.where(colors.reshape(-1, 1), (syn - meas) ** 2 / e_meas ** 2, (syn * scale - meas) ** 2 / e_meas ** 2)

    # Then add Chi2 of derived properties as distance, mass ratio, ...
    # =================================================================
    derived_properties = kwargs.get('derived_properties', {})
    constraints = kwargs.get('constraints', {})

    for con, (c, c_m, c_p) in list(constraints.items()):
        syn_c = None
        if con == 'distance' and sum(~colors) > 0:
            # distance can only be constrained if there is at least 1 absolute measurement
            c, c_m, c_p = 1. / c ** 2, 2. * c_m / c ** 3, 2. * c_p / c ** 3
            syn_c = scale

        elif con in pars:
            # constrained on a fitted parameter
            syn_c = pars[con]

        elif con in derived_properties:
            # constrained on a derived parameter
            syn_c = derived_properties[con]

        if not syn_c is None:
            chi2_c = np.where(syn_c < c, (syn_c - c) ** 2 / c_m ** 2, (syn_c - c) ** 2 / c_p ** 2)

            # append to chisq array
            chisq = np.vstack([chisq, chi2_c])

    return chisq.sum(axis=0), scale, e_scale



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "numpy",
    "emcee",
    "astropy",
    "astroquery",
    "pyvo",
    "matplotlib",
    "pyaml",
    "corner",
    "tqdm",
    "scipy",
    "h5py",
    "pandas",
]

setuptools.setup(
    name="speedyfit", # Replace with your own username
    version="0.0.3",
    author="Joris Vos",
    author_email="joris.vos@uv.cl",
    description="MC approach to fit photometric SEDs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vosjo/speedyfit",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    test_suite='pytest.collector',
    tests_require=['pytest'],
    entry_points={
        'console_scripts': ['speedyfit=speedyfit.speedyfit:main', 'speedyfit-batch=speedyfit.speedyfit_batch:main'],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    python_requires='>=3.7',
)



import numpy as np
from numpy.lib.recfunctions import merge_arrays

import emcee

from . import statfunc, model, filters

def lnlike(pars, derived_properties, y, yerr, **kwargs):
    """
    log likelihood function

    Calculates the chi2 of the model defined by theta compared to the observed magnitudes
    and colors. Will also take possible constraints on q, lr and d into account.
    """
    model_func = kwargs.pop('model_func', model.get_itable)
    stat_func = kwargs.pop('stat_func', statfunc.stat_chi2)
    colors = kwargs.get('colors', np.array([False for i in y], bool))
    constraints = kwargs.pop('constraints', {})


    #-- calculate synthetic magnitudes **kwargs contains info about which grid to use
    kwargs.update(pars)
    y_syn, extra_drv = model_func(**kwargs)
    if 'L' in extra_drv and 'L2' in extra_drv:
        #-- update light ratio with correct value from models
        extra_drv['lr'] = extra_drv['L'] / extra_drv['L2']
    derived_properties.update(extra_drv)


    chi2, scales, e_scales = stat_func(y,
                                       yerr,
                                       colors, y_syn, pars,
                                       derived_properties=derived_properties,
                                       constraints=constraints)

    #-- add distance to extra derived parameter (which already contains luminosities)
    #   distance is converted from Rsol to pc
    extra_drv['d'] = np.sqrt(1/scales)/44365810.04823812
    extra_drv['scale'] = scales
    extra_drv['chi2'] = chi2

    return -chi2/2, extra_drv

def lnprior(theta, derived_properties, limits, **kwargs):
    """
    Simple uniform (flat) prior on all parameters if the parameters
    are within their range, and the derived properties (q, m, ..) are also
    within their limits.

    if all parameters are within the provided limits, the the returned
    log probability is 0, otherwise it is -inf.

    :param theta: list of model parameters
    :type theta: list
    :param limits: limits on the model parameters
    :type limits: list of tuples

    :return: logarithm of the probability of the parameters (theta) given the
             model limits
    :rtype: float
    """

    derived_limits = kwargs.pop('derived_limits', {})

    #-- check if all parameters are within their limits
    if any(theta < limits[:,0]) or any(theta > limits[:,1]):
        return -np.inf

    #-- check that all derived properties are within limits
    for lim in list(derived_limits.keys()):
        if derived_properties[lim] < derived_limits[lim][0] or \
                derived_properties[lim] > derived_limits[lim][1]:
            return -np.inf

    return 0

def lnprob(theta, y, yerr, limits, **kwargs):
    """
    full log probability function combining the prior and the likelihood

    will return -inf if any of :py:func:`lnprior` or :py:func:`lnlikelyhood` is
    infite, otherwise it will return the sum of both functions.

    :param theta: list of model parameters (normaly mass, fe/h and age)
    :type theta: list
    :param y: 1D array of observables
    :type y: array
    :param yerr: 1D array containing errors on every observable
    :type yerr: array
    :param limits: limits on the model parameters
    :type limits: list of tuples

    :return: the sum of the log prior and log likelihood
    :rtype: float
    """

    #-- create keyword parameters from theta
    pars = {}
    for name, value in zip(kwargs['pnames'], theta):
        pars[name]=value

    #-- add extra variables which are not fitted to pars.
    pars.update(kwargs.pop('fixed_variables', {}))

    #-- get derived properties
    prop_func = kwargs.pop('prop_func', statfunc.get_derived_properties)
    syn_drv = prop_func(**pars)

    if 'rad' in syn_drv:
        pars['rad']=syn_drv['rad']
    if 'rad2' in syn_drv:
        pars['rad2']=syn_drv['rad2']

    #-- calculate prior probability
    lp = lnprior(theta, syn_drv, limits, **kwargs)
    if not np.isfinite(lp):
        return -np.inf, syn_drv

    #-- calculate likelihood
    ll, extra_drv = lnlike(pars, syn_drv, y, yerr, **kwargs)
    syn_drv.update(extra_drv)
    if not np.isfinite(ll):
        return -np.inf, syn_drv

    return lp + ll, syn_drv


def MCMC(obs, obs_err, photbands,
         pnames, limits, grids,
         fixed_variables={}, constraints={}, derived_limits={},
         nwalkers=100, nsteps=1000, nrelax=150, a=10, pos=None):

    #-- check which bands are colors
    colors = np.array([filters.is_color(photband) for photband in photbands], bool)

    #-- initialize the walkers randomly is no starting positions are given
    if pos is None:
        pos = [ np.random.uniform(lim[0], lim[1], nwalkers) for lim in limits]
        pos = np.array(pos).T
    else:
        nwalkers = pos.shape[0]

    #-- setup the sampler
    ndim = len(pnames)
    kwargs = {'pnames':pnames,
              'colors':colors,
              'grid':grids,
              'fixed_variables':fixed_variables,
              'constraints':constraints,
              'derived_limits':derived_limits,
              'prop_func': statfunc.get_derived_properties}

    # TODO: storing the blobs as dictionary with dtype object and then later converting to recarray is inefficient.
    # This needs to be addressed: provide the correct dtypes here and let emcee directly store them in recarray
    sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob, a=a,
                                    args=(obs, obs_err, limits), kwargs=kwargs, blobs_dtype=[('blob','O')])

    #================
    # MCMC part

    #-- run the sampler, both burn in and actual run in one
    sampler.run_mcmc(pos, nsteps+nrelax, progress=True)

    #-- combine the results from the individual walkers discarding the burn in steps
    samples = sampler.get_chain(discard=nrelax, thin=1, flat=True)
    blobs = sampler.get_blobs(discard=nrelax, thin=1, flat=True)
    probabilities = sampler.get_log_prob(discard=nrelax, thin=1, flat=True)

    #-- clear the samples to save memory
    sampler.reset()

    #-- remove all steps that are not accepted (lnprob == -inf)
    accept = np.where(np.isfinite(probabilities))
    samples = samples[accept]
    blobs = blobs[accept]
    probabilities = probabilities[accept]

    if len(samples) == 0:
        raise ValueError('No models were accepted, all probabilities were -inf')

    #-- convert to recarrays
    dtypes = [(n, 'f8') for n in pnames]
    samples = np.array([tuple(s) for s in samples], dtype=dtypes)

    names = list(blobs[0][0].keys())
    pars = []
    for b in blobs:
        pars.append(tuple([b[0][n] for n in names]))
    dtypes = [(n, 'f8') for n in names]
    blobs = np.array(pars, dtype=dtypes)

    #-- remove all steps where model creation failed (d == 0)
    accept = np.where(blobs['d'] > 0)
    samples = samples[accept]
    blobs = blobs[accept]
    probabilities = probabilities[accept]

    #-- merge all results in 1 recarray and select best model
    data = merge_arrays((samples, blobs), asrecarray=True, flatten=True)
    best = np.where(probabilities == np.max(probabilities))

    results = {}
    for n, v in zip(data.dtype.names, data[best][0]):
        results[n] = v


    return results, data
   
   

   
   



import re
import os
import numpy as np
 
from astropy.io import fits 

from . import interpol, filters

__defaults__ = dict(grid='kurucz2',
                    directory=os.environ.get('SPEEDYFIT_MODELS', None),)
defaults = __defaults__.copy()

def check_grids():

   print("Checking which atmosphere models are available...")

   if defaults['directory'] is not None:
      print("Checking for models in {}".format(defaults['directory']))
   else:
      print("SPEEDYFIT_MODELS environmental variable not set. CAN NOT find models!")
      print("Please point the SPEEDYFIT_MODELS variable to the directory where you stored models. On bash use:")
      print("\texport SPEEDYFIT_MODELS='<path to extracted atmosphere models>'")
      return

   for grid in ['kurucz2', 'munari', 'tmap', 'koester', 'blackbody']:
      print(grid)
      gridpath = get_grid_file(integrated=False, grid=grid)
      if os.path.isfile(gridpath):
         print("\t raw: available")
      else:
         print("\t raw: NOT FOUND")

      gridpath = get_grid_file(integrated=True, grid=grid)
      if os.path.isfile(gridpath):
         print("\t integrated: available")
      else:
         print("\t integrated: NOT FOUND")


def get_grid_file(integrated=False, **kwargs):
   
   grid = kwargs.get('grid', defaults['grid'])
   
   if os.path.isfile(grid):
      return grid
   
   if grid == 'kurucz2':
      filename = 'kurucz93_z0.0_k2odfnew_sed'
      
   elif grid == 'munari':
      filename = 'Munari2005_extended'
      
   elif grid == 'tmap':
      filename = 'TMAP2012_sdOB_extended'
   
   elif grid == 'koester':
      filename = 'Koester_WD_DA_2014_extended'
   
   elif grid == 'blackbody':
      filename = 'blackbody_discint'
   
   else:
      raise ValueError('Grid name ({}) not recognized!'.format(grid))
      
   if integrated:
      filename = 'i' + filename + '_lawfitzpatrick2004_Rv3.10'
   
   directory = kwargs.get('directory', defaults['directory'])
   
   return directory + filename + '.fits'


def load_grids(gridnames, pnames, limits, photbands):
   """
   prepares the integrated photometry grid by loading the grid and cutting it to the size
   given in limits.
   """
   grids = []
   for i, name in enumerate(gridnames):
      
      ind = '' if i == 0 else str(i + 1)
      
      axis_values, grid_pars, pixelgrid, grid_names = prepare_grid(photbands, name,
            teffrange=limits[pnames.index('teff'+ind)],
            loggrange=limits[pnames.index('logg'+ind)],
            ebvrange =limits[pnames.index('ebv')],
            variables=['teff','logg','ebv'])

      grids.append([axis_values, pixelgrid])
      
   return grids


def prepare_grid(photbands, gridname,
                 teffrange=(-np.inf,np.inf),loggrange=(-np.inf,np.inf),
                 ebvrange=(-np.inf,np.inf),
                 variables=['teff','logg','ebv'],**kwargs):
   
   flux = []
   grid_pars = []
   grid_names = np.array(variables)
   
   gridfilename = get_grid_file(integrated=True, grid=gridname)
   
   with fits.open(gridfilename) as ff:
      #-- make an alias for further reference
      ext = ff[1]
      
      #-- the grid is already cut here to limit memory usage
      keep = np.ones(len(ext.data),bool)
      for name in variables:
         #-- first find the closest actuall grid points
         low,high = locals()[name+'range']
         lidx = np.abs(ext.data.field(name)[ext.data.field(name) <= low]-low).argmin()
         hidx = np.abs(ext.data.field(name)[ext.data.field(name) >= high]-high).argmin()
         low = ext.data.field(name)[ext.data.field(name) <= low][lidx]
         high = ext.data.field(name)[ext.data.field(name) >= high][hidx]
         
         #-- we need to be carefull for rounding errors
         in_range = (low<=ext.data.field(name)) & (ext.data.field(name)<=high)
         on_edge  = np.allclose(ext.data.field(name),low) | np.allclose(ext.data.field(name),high)
         
         keep = keep & (in_range | on_edge)
         
      grid_pars = np.vstack([ext.data.field(name)[keep] for name in variables])
      flux = _get_flux_from_table(ext,photbands,include_Labs=True)[keep]
      
   flux = np.log10(flux)
   
   #-- create the pixeltype grid
   axis_values, pixelgrid = interpol.create_pixeltypegrid(grid_pars, flux.T)
   return axis_values,grid_pars.T,pixelgrid,grid_names

def get_itable_single(teff=None, logg=None, g=None, ebv=0.0, **kwargs):
   
   #-- check if logg or g is given
   if logg is None and not g is None:
      logg = np.log10(g)
   
   multiple = False
   if not hasattr(teff, '__iter__') or not hasattr(logg, '__iter__') or \
      not hasattr(ebv, '__iter__'):
      teff = np.array(teff)
      logg = np.array(logg)
      ebv = np.array(ebv)
      multiple = True
   
   #-- get the grid from the keyword, or prepare it if a grid name is given
   if isinstance(kwargs['grid'], str):
      axis_values,grid_pars,pixelgrid,grid_names = prepare_grid(kwargs['photbands'], kwargs['grid'],
                 teffrange=(np.min(teff), np.max(teff)),loggrange=(np.min(logg), np.max(logg)),
                 ebvrange=(np.min(ebv), np.max(ebv)),
                 variables=['teff','logg','ebv'])
   else:
      axis_values, pixelgrid = kwargs['grid']
   
   
   
   p = np.vstack([teff, logg, ebv])
   
   values = interpol.interpolate(p, axis_values, pixelgrid)
      
   #-- switch logarithm to normal
   values = 10**values
   flux,Labs = values[:-1],values[-1]
   
   #-- Take radius into account when provided
   if 'rad' in kwargs:
      rad = np.array(kwargs['rad'])
      flux,Labs = flux*rad**2, Labs*rad**2
   
   if 'd' in kwargs:
      d = np.array(kwargs['d'])
      flux,Labs = flux/d**2, Labs*d**2
   
   if multiple:
      flux = flux.flatten()
      Labs = Labs.flatten()
   
   return flux, Labs
   

def get_itable(grid=[], **kwargs):
   
   values, parameters, components = {}, set(), set()
   for key in list(kwargs.keys()):
      if re.search("^(teff|logg|g|ebv|rad)\d?$", key):
         par, comp = re.findall("^(teff|logg|g|ebv|rad)(\d?)$", key)[0]
         values[key] = kwargs.pop(key)
         parameters.add(par)
         components.add(comp)
   
   # need to check here that grid is same length as components
   if isinstance(grid, str):
      grids = [grid]
   elif hasattr(grid, '__iter__') and isinstance(grid[0], str):
      grids = grid
   elif hasattr(grid, '__iter__') and  hasattr(grid[0], '__iter__') and len(grid[0]) == 2:
      grids = grid
   else:
      grids = [grid]
   
   #-- If there is only one component, we can directly return the result
   if len(components) == 1:
      kwargs.update(values)
      fluxes, Labs = get_itable_single(grid=grids[0], **kwargs)
      return fluxes, {'L':Labs}
   
   fluxes, Labs = [],{}                               
   for i, (comp, grid) in enumerate(zip(components,grids)):
      kwargs_ = kwargs.copy()
      for par in parameters:
         kwargs_[par] = values[par+comp] if par+comp in values else values[par]
      
      f,L = get_itable_single(grid=grid, **kwargs_)
                                    
      fluxes.append(f)
      Labs['L'+comp] = np.sum(L,axis=0)
   
   fluxes = np.sum(fluxes,axis=0)
   return fluxes, Labs

def get_table_single(teff=None, logg=None, ebv=0.0, **kwargs):
   """
   No interpolating, just returns the closest gridpoint
   """
   
   #-- get the grid
   gridname = kwargs['grid']
   gridfilename = get_grid_file(integrated=False, grid=gridname)
   
   hdu = fits.open(gridfilename)
   
   teffs, loggs = np.zeros(len(hdu)-1), np.zeros(len(hdu)-1)
   for i in range(1,len(hdu)):
      teffs[i-1] = hdu[i].header['TEFF']
      loggs[i-1] = hdu[i].header['LOGG']
   
   dteff = abs(teffs-teff)/teff
   dlogg = abs(loggs-logg)/logg
   
   s = np.where(np.sqrt(dteff**2 + dlogg**2) == np.min(np.sqrt(dteff**2 + dlogg**2)))
   
   model = hdu[s[0][0]+1].data
   wave, flux = model['wavelength'], model['flux']
   
   #-- Take radius into account when provided
   if 'rad' in kwargs:
      rad = np.array(kwargs['rad'])
      flux = flux*rad**2
      
   if 'd' in kwargs:
      d = np.array(kwargs['d'])
      flux = flux/d**2
   
   return wave, flux
   
def get_table(grid=[], **kwargs):
   """
   Returns the closest model atmosphere available in the grid. No interpolation is done!
   """
   values, parameters, components = {}, set(), set()
   for key in list(kwargs.keys()):
      if re.search("^(teff|logg|g|ebv|rad)\d?$", key):
         par, comp = re.findall("^(teff|logg|g|ebv|rad)(\d?)$", key)[0]
         values[key] = kwargs.pop(key)
         parameters.add(par)
         components.add(comp)
   
   # need to check here that grid is same length as components
   if hasattr(grid, '__iter__') and not isinstance(grid, str):
      grids = grid
   else:
      grids = [grid]
   
   #-- If there is only one component, we can directly return the result
   if len(components) == 1:
      kwargs.update(values)
      wave, flux = get_table_single(grid=grids[0], **kwargs)
      return wave, flux
   
   waves, fluxes = [], []
   for i, (comp, grid) in enumerate(zip(components,grids)):
      kwargs_ = kwargs.copy()
      for par in parameters:
         kwargs_[par] = values[par+comp] if par+comp in values else values[par]
      
      w, f = get_table_single(grid=grid, **kwargs_)
      
      waves.append(w)
      fluxes.append(f)
   
   #-- interpolate and combine the models
   wave = waves[0]
   flux = np.zeros_like(wave)
   for w, f in zip(waves, fluxes):
      flux += np.interp(wave, w, f)
   
   return wave, flux
   

def _get_flux_from_table(fits_ext,photbands,index=None,include_Labs=True):
   """
   Retrieve flux and flux ratios from an integrated SED table.
   
   @param fits_ext: fits extension containing integrated flux
   @type fits_ext: FITS extension
   @param photbands: list of photometric passbands
   @type photbands: list of str
   @param index: slice or index of rows to retrieve
   @type index: slice or integer
   @return: fluxes or flux ratios
   #@rtype: list
   """
   if index is None:
      index = slice(None) #-- full range
   fluxes = []
   for photband in photbands:
      try:
         if not filters.is_color(photband):
            fluxes.append(fits_ext.data.field(photband)[index])
         else:
            system,color = photband.split('.')
            if '-' in color:
               band0,band1 = color.split('-')
               fluxes.append(fits_ext.data.field('%s.%s'%(system,band0))[index]/fits_ext.data.field('%s.%s'%(system,band1))[index])
            elif color=='M1':
               fv = fits_ext.data.field('STROMGREN.V')[index]
               fy = fits_ext.data.field('STROMGREN.Y')[index]
               fb = fits_ext.data.field('STROMGREN.B')[index]
               fluxes.append(fv*fy/fb**2)
            elif color=='C1':
               fu = fits_ext.data.field('STROMGREN.U')[index]
               fv = fits_ext.data.field('STROMGREN.V')[index]
               fb = fits_ext.data.field('STROMGREN.B')[index]
               fluxes.append(fu*fb/fv**2)
      except KeyError:
         print('Passband %s missing from table'%(photband))
         fluxes.append(np.nan*np.ones(len(fits_ext.data)))
   #-- possibly include absolute luminosity
   if include_Labs:
      fluxes.append(fits_ext.data.field("Labs")[index])
   fluxes = np.array(fluxes).T
   if index is not None:
      fluxes = fluxes
   return fluxes





import os

import configparser

import numpy as np

from . import filters

from astroquery.simbad import Simbad
from astroquery.vizier import Vizier
from astroquery.utils.tap.core import TapPlus
import pyvo

from astropy import units as u
from astropy.coordinates.angles import Angle
from astropy.coordinates import SkyCoord
from astropy.io import ascii

from numpy.lib.recfunctions import append_fields

filedir = os.path.dirname(os.path.abspath(__file__))

# always request the distance of the observations to the search coordinates and sort on increasing distance
v = Vizier(columns=["*", '+_r']) 

#-- read in catalog information
viz_info = configparser.ConfigParser()
viz_info.optionxform = str # make sure the options are case sensitive
viz_info.readfp(open(filedir+'/vizier_cats_phot.cfg'))


tap_info = configparser.ConfigParser()
tap_info.optionxform = str # make sure the options are case sensitive
tap_info.readfp(open(filedir+'/tap_cats_phot.cfg'))


def get_coordinate(objectname):
   
   if objectname[0] == 'J' and objectname[1] != 'L':
      name = objectname[1:]
      if '+' in name:
         ra, dec = name.split('+')
         sign = '+'
      else:
         ra, dec = name.split('-')
         sign = '-'
      
      ra = "{}:{}:{}".format(ra[0:2], ra[2:4], ra[4:])
      dec = "{}{}:{}:{}".format(sign, dec[0:2], dec[2:4], dec[4:])
      
      ra = Angle(ra, unit='hour').degree
      dec = Angle(dec, unit='degree').degree
      
   else:
      data = Simbad.query_object(objectname)
      ra = Angle(data['RA'][0], unit='hour').degree
      dec = Angle(data['DEC'][0], unit='degree').degree
      
   return ra, dec


def get_vizier_photometry(objectname, radius=5):
   
   catalogs = viz_info.sections()

   data = v.query_object(objectname, catalog=catalogs, radius=radius*u.arcsec)
   
   photometry = []
   
   for catalog in list(data.keys()):
      
      distance = (data[catalog]['_r'][0] * u.arcmin).to(u.arcsec).value
      
      if viz_info.has_option(catalog, 'bibcode'):
         bibcode = viz_info.get(catalog, 'bibcode')
      else:
         bibcode = '-'
      
      print(catalog)
      
      for band in viz_info.options(catalog):
         
         if '_unit' in band: continue
         if '_err' in band: continue
         if 'bibcode' in band: continue
         if '-' in band: continue
         
         bandname = viz_info.get(catalog, band)
         value = data[catalog][band][0]
         
         errkw = viz_info.get(catalog, band+'_err') if viz_info.has_option(catalog, band+'_err') else 'e_'+band
         try:
            err = data[catalog][errkw][0]
         except:
            err = 0.05
         
         if viz_info.has_option(catalog, band+'_unit'):
            unit = viz_info.get(catalog, band+'_unit')
         else:
            unit = data[catalog][band].unit
         
         photometry.append(( bandname, value, err, unit, distance, bibcode ))
   
   dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
   photometry = np.array(photometry, dtype=dtypes)
   
   return photometry


def tap_query_vo(ra, dec, catalog):
   """
   info see:
   http://docs.g-vo.org/pyvo/html/page011.html
   """

   service = pyvo.dal.TAPService(catalog)

   table = tap_info.get(catalog, 'table')
   print(table)

   keywords = ""
   bands = []
   for band in tap_info.options(catalog):
      if 'table' in band: continue
      if 'rakw' in band: continue
      if 'deckw' in band: continue
      if '_unit' in band: continue
      if '_err' in band: continue
      if 'bibcode' in band: continue

      errkw = tap_info.get(catalog, band + '_err') if tap_info.has_option(catalog, band + '_err') else 'e_' + band

      keywords += band + ", " + errkw + ", "
      bands.append(band)

   if keywords[-2:] == ", ":
      keywords = keywords[0:-2]

   rakw = tap_info.get(catalog, 'rakw') if tap_info.has_option(catalog, 'rakw') else 'raj2000'
   deckw = tap_info.get(catalog, 'deckw') if tap_info.has_option(catalog, 'deckw') else 'dej2000'

   # -- first try to query with distance
   query = """SELECT 
         DISTANCE(POINT('ICRS', {rakw:}, {deckw}),
                  POINT('ICRS', {ra:}, {dec:})) AS dist, {kws:}
         FROM {table:} AS m
         WHERE 
            1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                     CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
         ORDER BY dist""".format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)

   results = service.run_sync(query).to_table()

   if len(results) == 0:
      dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
      return np.array([], dtype=dtypes)

   # -- get the distance from the query result or calculate it.
   if 'dist' in results.dtype.names:
      distance = (results['dist'][0] * u.degree).to(u.arcsec).value
   else:
      c1 = SkyCoord(ra * u.degree, dec * u.degree)
      distance = SkyCoord(results['ra'][0] * u.degree, results['de'][0] * u.degree).separation(c1).to(u.arcsec).value

   bibcode = tap_info.get(catalog, 'bibcode')

   # -- get the photometric measurements, errors and units
   photometry = []
   for band in bands:
      bandname = tap_info.get(catalog, band)
      value = results[band][0]

      errkw = tap_info.get(catalog, band + '_err') if tap_info.has_option(catalog, band + '_err') else 'e_' + band
      err = results[errkw][0]

      if tap_info.has_option(catalog, band + '_unit'):
         unit = tap_info.get(catalog, band + '_unit')
      else:
         unit = results[band].unit

      photometry.append((bandname, value, err, unit, distance, bibcode))

   dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
   photometry = np.array(photometry, dtype=dtypes)

   return photometry


def tap_query(ra, dec, catalog):
   
   tp = TapPlus(url=catalog)
   
   table = tap_info.get(catalog, 'table')
   print(table)
   
   keywords = ""
   bands = []
   for band in tap_info.options(catalog):
      if 'table' in band: continue
      if 'rakw' in band: continue
      if 'deckw' in band: continue
      if '_unit' in band: continue
      if '_err' in band: continue
      if 'bibcode' in band: continue
      
      errkw = tap_info.get(catalog, band+'_err') if tap_info.has_option(catalog, band+'_err') else 'e_'+band

      keywords += band + ", " + errkw + ", "
      bands.append(band)
      
   if keywords[-2:] == ", ":
      keywords = keywords[0:-2]
      
   rakw = tap_info.get(catalog, 'rakw') if tap_info.has_option(catalog, 'rakw') else 'raj2000'
   deckw = tap_info.get(catalog, 'deckw') if tap_info.has_option(catalog, 'deckw') else 'dej2000'
   
   try:
      #-- first try to query with distance
      query = """SELECT 
      DISTANCE(POINT('ICRS', {rakw:}, {deckw}),
               POINT('ICRS', {ra:}, {dec:})) AS dist, {kws:}
      FROM {table:} AS m
      WHERE 
         1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                  CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
      ORDER BY dist""".format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)
      
      job = tp.launch_job(query) 
      
      results = job.get_results()
      
   except Exception as e:
      
      print("problem", e)
      
      #-- not all catalogs accept a distance sorted query, we have to hope that the correct star is returned.
      query = """SELECT {rakw:} as RA, {deckw} as DE, {kws:}
      FROM {table:} AS m
      WHERE 
         1=CONTAINS(POINT('ICRS', {rakw:}, {deckw}),
                  CIRCLE('ICRS', {ra:}, {dec:}, 0.005 ))
      """.format(ra=ra, dec=dec, table=table, rakw=rakw, deckw=deckw, kws=keywords)
      
      job = tp.launch_job(query) 
      
      results = job.get_results()
   
   if len(results) == 0:
      dtypes = [('band', 'a20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', 'a10'), ('distance', 'f8'), ('bibcode', 'a20')]
      return np.array([], dtype=dtypes)
   
   #-- get the distance from the query result or calculate it.
   if 'dist' in results.dtype.names:
      distance = (results['dist'][0] * u.degree).to(u.arcsec).value
   else:
      c1 = SkyCoord(ra * u.degree, dec * u.degree)
      distance = SkyCoord(results['ra'][0] * u.degree, results['de'][0] * u.degree).separation(c1).to(u.arcsec).value
      
      
   bibcode = tap_info.get(catalog, 'bibcode')
   
   #-- get the photometric measurements, errors and units
   photometry = []
   for band in bands:
      bandname = tap_info.get(catalog, band)
      value = results[band][0]
      
      errkw = tap_info.get(catalog, band+'_err') if tap_info.has_option(catalog, band+'_err') else 'e_'+band
      err = results[errkw][0]
      
      if tap_info.has_option(catalog, band+'_unit'):
         unit = tap_info.get(catalog, band+'_unit')
      else:
         unit = results[band].unit
      
      photometry.append(( bandname, value, err, unit, distance, bibcode))
      
   dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
   photometry = np.array(photometry, dtype=dtypes)

   return photometry


def get_tap_photometry(ra, dec):
   catalogs = tap_info.sections()
   
   dtypes = [('band', '<U20'), ('meas', 'f8'), ('emeas', 'f8'), ('unit', '<U10'), ('distance', 'f8'), ('bibcode', '<U20')]
   photometry = np.array([], dtype=dtypes)
   
   for catalog in catalogs:
      
      phot_ = tap_query_vo(ra, dec, catalog)
      photometry = np.hstack([photometry, phot_])
      
   return photometry


def get_photometry(objectname, filename=None):
   
   #-- get the object coodinates
   ra, dec = get_coordinate(objectname)
   
   #-- query tap catalogs
   photometry = get_tap_photometry(ra, dec)
   
   #-- query Vizier catalogs
   photometry_ = get_vizier_photometry("{} {}".format(ra, dec))

   print(photometry)
   print(photometry_)

   photometry = np.hstack([photometry, photometry_])

   #-- convert magnitudes to fluxes
   wave, flux, err = [], [], []
   for band, meas, emeas, unit in zip(photometry['band'], photometry['meas'], photometry['emeas'], photometry['unit']):
      if np.isnan(emeas) or emeas < 0: emeas = 0.02

      f_, e_ = filters.mag2flux(meas, emeas, band)
      
      wave.append(filters.eff_wave(band))
      flux.append(f_)
      err.append(e_)
   
   photometry = append_fields(photometry, ['flux', 'eflux'], data=[flux, err], 
                              dtypes=['f8', 'f8'], usemask=False)
   
   if filename is not None:
      ascii.write(photometry, filename, format='fixed_width', overwrite=True)
   
   return photometry


def get_parallax(objectname, radius=5):
   
   #data = Simbad.query_object(objectname)
   #ra = Angle(data['RA'][0], unit='hour').degree
   #dec = Angle(data['DEC'][0], unit='degree').degree
   
   v_gaia = Vizier(columns=["Plx", "e_Plx", '+_r']) 
      
   data = v_gaia.query_object(objectname, catalog=['I/345/gaia2'], radius=radius*u.arcsec)

   if len(data) == 0:
      return None, None

   return data['I/345/gaia2']['Plx'][0], data['I/345/gaia2']['e_Plx'][0]
   

#photometry = get_photometry('SB 705', 'SB_705.phot')

#print photometry


#get_parallax('SB 705', radius=5)



# -*- coding: utf-8 -*-

import os
import numpy as np

from . import filters
from . import model
from . import fileio

basename = os.path.join(os.path.dirname(__file__),'redlaws')

#{ Main interface

def get_law(name,norm='E(B-V)',photbands=None,**kwargs):
    """
    Retrieve an interstellar reddening law.
    
    Parameter C{name} must be the function name of one of the laws defined in
    this module.
    
    By default, the law will be interpolated on a grid from 100 angstrom to
    10 micron in steps of 10 angstrom. This can be adjusted with the parameter
    C{wave} (array), which B{must} be in angstrom. You can change the units
    ouf the returned wavelength array via C{wave_units}.
    
    By default, the curve is normalised with respect to E(B-V) (you get
    A(l)/E(B-V)). You can set the C{norm} keyword to Av if you want A(l)/Av.
    Remember that
    
    A(V) = Rv * E(B-V)
    
    The parameter C{Rv} is by default 3.1, other reasonable values lie between
    2.0 and 5.1
    
    Extra accepted keywords depend on the type of reddening law used.
    
    Example usage:
    
    >>> wave = np.r_[1e3:1e5:10]
    >>> wave,mag = get_law('cardelli1989',wave=wave,Rv=3.1)
    
    @param name: name of the interstellar law
    @type name: str, one of the functions defined here
    @param norm: type of normalisation of the curve
    @type norm: str (one of E(B-V), Av)
    @param photbands: list of photometric passbands
    @type photbands: list of strings
    @keyword wave: wavelength array to interpolate the law on
    @type wave: ndarray
    @return: wavelength, reddening magnitude
    @rtype: (ndarray,ndarray)
    """
    #-- get the inputs
    wave_ = kwargs.pop('wave',None)
    Rv = kwargs.setdefault('Rv',3.1)
    
    #-- get the curve
    wave,mag = globals()[name.lower()](**kwargs)
    wave_orig,mag_orig = wave.copy(),mag.copy()
    
    #-- interpolate on user defined grid
    if wave_ is not None:
        mag = np.interp(wave_,wave,mag,right=0)
        wave = wave_
           
    #-- pick right normalisation: convert to A(lambda)/Av if needed
    if norm.lower()=='e(b-v)':
        mag *= Rv
    else:
        #-- we allow ak and av as shortcuts for normalisation in JOHNSON K and
        #   V bands
        if norm.lower()=='ak':
            norm = 'JOHNSON.K'
        elif norm.lower()=='av':
            norm = 'JOHNSON.V'
        norm_reddening = model.synthetic_flux(wave_orig, mag_orig, [norm])[0]
        mag /= norm_reddening
    
    #-- maybe we want the curve in photometric filters
    if photbands is not None:
        mag = model.synthetic_flux(wave, mag, photbands)
        wave = filters.get_info(photbands)['eff_wave']
    
    return wave,mag


def redden(flux,wave=None,photbands=None,ebv=0.,rtype='flux',law='cardelli1989',**kwargs):
    """
    Redden flux or magnitudes
    
    The reddening parameters C{ebv} means E(B-V).
    
    If it is negative, we B{deredden}.
    
    If you give the keyword C{wave}, it is assumed that you want to (de)redden
    a B{model}, i.e. a spectral energy distribution.
    
    If you give the keyword C{photbands}, it is assumed that you want to (de)redden
    B{photometry}, i.e. integrated fluxes.
    
    @param flux: fluxes to (de)redden (magnitudes if C{rtype='mag'})
    @type flux: ndarray (floats)
    @param wave: wavelengths matching the fluxes (or give C{photbands})
    @type wave: ndarray (floats)
    @param photbands: photometry bands matching the fluxes (or give C{wave})
    @type photbands: ndarray of str
    @param ebv: reddening parameter E(B-V)
    @type ebv: float
    @param rtype: type of dereddening (magnituds or fluxes)
    @type rtype: str ('flux' or 'mag')
    @return: (de)reddened flux/magnitude
    @rtype: ndarray (floats)
    """
    if photbands is not None:
        wave = filters.get_info(photbands)['eff_wave']
        
    old_settings =  np.seterr(all='ignore')
    wave, reddeningMagnitude = get_law(law,wave=wave,**kwargs)

    if rtype=='flux':
        # In this case flux means really flux
        flux_reddened = flux / 10**(reddeningMagnitude*ebv/2.5)
        np.seterr(**old_settings)
        return flux_reddened
    elif rtype=='mag':
        # In this case flux means actually a magnitude
        magnitude = flux
        magnitude_reddened = magnitude + reddeningMagnitude*ebv
        np.seterr(**old_settings)
        return magnitude_reddened

def deredden(flux,wave=None,photbands=None,ebv=0.,rtype='flux',**kwargs):
    """
    Deredden flux or magnitudes.
    
    @param flux: fluxes to (de)redden (NOT magnitudes)
    @type flux: ndarray (floats)
    @param wave: wavelengths matching the fluxes (or give C{photbands})
    @type wave: ndarray (floats)
    @param photbands: photometry bands matching the fluxes (or give C{wave})
    @type photbands: ndarray of str
    @param ebv: reddening parameter E(B-V)
    @type ebv: float
    @param rtype: type of dereddening (magnituds or fluxes)
    @type rtype: str ('flux' or 'mag')
    @return: (de)reddened flux
    @rtype: ndarray (floats)
    """
    return redden(flux,wave=wave,photbands=photbands,ebv=-ebv,rtype=rtype,**kwargs)
    

#}

#{ Curve definitions
def chiar2006(Rv=3.1,curve='ism',**kwargs):
    """
    Extinction curve at infrared wavelengths from Chiar and Tielens (2006)
    
    We return A(lambda)/E(B-V), by multiplying A(lambda)/Av with Rv.
    
    This is only defined for Rv=3.1. If it is different, this will raise an
    AssertionError
    
    Extra kwags are to catch unwanted keyword arguments.
    
    UNCERTAIN NORMALISATION
    
    @param Rv: Rv
    @type Rv: float
    @param curve: extinction curve
    @type curve: string (one of 'gc' or 'ism', galactic centre or local ISM)
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    source = os.path.join(basename,'Chiar2006.red')
    
    #-- check Rv
    assert(Rv==3.1)
    wavelengths,gc,ism = fileio.read2array(source).T
    if curve=='gc':
        alam_ak = gc
    elif curve=='ism':
        keep = ism>0
        alam_ak = ism[keep]
        wavelengths = wavelengths[keep]
    else:
        raise ValueError('no curve %s'%(curve))
    alam_aV = alam_ak * 0.09
    #plot(1/wavelengths,alam_aV,'o-')
    return wavelengths*1e4,alam_aV



def fitzpatrick1999(Rv=3.1,**kwargs):
    """
    From Fitzpatrick 1999 (downloaded from ASAGIO database)
    
    This function returns A(lambda)/A(V).
    
    To get A(lambda)/E(B-V), multiply the return value with Rv (A(V)=Rv*E(B-V))
    
    Extra kwags are to catch unwanted keyword arguments.
    
    @param Rv: Rv (2.1, 3.1 or 5.0)
    @type Rv: float
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    filename = 'Fitzpatrick1999_Rv_%.1f'%(Rv)
    filename = filename.replace('.','_') + '.red'
    myfile = os.path.join(basename,filename)
    wave,alam_ebv = fileio.read2array(myfile).T
    alam_av = alam_ebv/Rv
    
    return wave,alam_av

def fitzpatrick2004(Rv=3.1,**kwargs):
    """
    From Fitzpatrick 2004 (downloaded from FTP)
    
    This function returns A(lambda)/A(V).
    
    To get A(lambda)/E(B-V), multiply the return value with Rv (A(V)=Rv*E(B-V))
    
    Extra kwags are to catch unwanted keyword arguments.
    
    @param Rv: Rv (2.1, 3.1 or 5.0)
    @type Rv: float
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    filename = 'Fitzpatrick2004_Rv_%.1f.red'%(Rv)
    myfile = os.path.join(basename,filename)
    wave_inv,elamv_ebv = fileio.read2array(myfile, skip_lines=15).T
    
    return 1e4/wave_inv[::-1],((elamv_ebv+Rv)/Rv)[::-1]


def donnell1994(**kwargs):
    """
    Small improvement on Cardelli 1989 by James E. O'Donnell (1994).
    
    Extra kwags are to catch unwanted keyword arguments.
    
    @keyword Rv: Rv
    @type Rv: float
    @keyword wave: wavelengths to compute the curve on
    @type wave: ndarray
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    return cardelli1989(curve='donnell',**kwargs)


def cardelli1989(Rv=3.1,curve='cardelli',wave=None,**kwargs):
    """
    Construct extinction laws from Cardelli (1989).
    
    Improvement in optical by James E. O'Donnell (1994)
    
    wavelengths in Angstrom!
    
    This function returns A(lambda)/A(V).
    
    To get A(lambda)/E(B-V), multiply the return value with Rv (A(V)=Rv*E(B-V))
    
    Extra kwags are to catch unwanted keyword arguments.
    
    @param Rv: Rv
    @type Rv: float
    @param curve: extinction curve
    @type curve: string (one of 'cardelli' or 'donnell')
    @param wave: wavelengths to compute the curve on
    @type wave: ndarray
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    if wave is None:
        wave = np.r_[100.:100000.:10]
    
    all_x = 1./(wave/1.0e4)
    alam_aV = np.zeros_like(all_x)
    
    #-- infrared
    infrared = all_x<1.1
    x = all_x[infrared]
    ax = +0.574*x**1.61
    bx = -0.527*x**1.61
    alam_aV[infrared] = ax + bx/Rv
    
    #-- optical
    optical = (1.1<=all_x) & (all_x<3.3)
    x = all_x[optical]
    y = x-1.82
    if curve=='cardelli':
        ax = 1 + 0.17699*y    - 0.50447*y**2 - 0.02427*y**3 + 0.72085*y**4 \
               + 0.01979*y**5 - 0.77530*y**6 + 0.32999*y**7
        bx =     1.41338*y    + 2.28305*y**2 + 1.07233*y**3 - 5.38434*y**4 \
               - 0.62251*y**5 + 5.30260*y**6 - 2.09002*y**7
    elif curve=='donnell':
        ax = 1 + 0.104*y    - 0.609*y**2 + 0.701*y**3 + 1.137*y**4 \
               - 1.718*y**5 - 0.827*y**6 + 1.647*y**7 - 0.505*y**8
        bx =     1.952*y    + 2.908*y**2 - 3.989*y**3 - 7.985*y**4 \
              + 11.102*y**5 + 5.491*y**6 -10.805*y**7 + 3.347*y**8
    else:
        raise ValueError('curve %s not found'%(curve))
    alam_aV[optical] = ax + bx/Rv
    
    #-- ultraviolet
    ultraviolet = (3.3<=all_x) & (all_x<8.0)
    x = all_x[ultraviolet]
    Fax = -0.04473*(x-5.9)**2 - 0.009779*(x-5.9)**3
    Fbx = +0.21300*(x-5.9)**2 + 0.120700*(x-5.9)**3
    Fax[x<5.9] = 0
    Fbx[x<5.9] = 0
    ax = +1.752 - 0.316*x - 0.104 / ((x-4.67)**2 + 0.341) + Fax
    bx = -3.090 + 1.825*x + 1.206 / ((x-4.62)**2 + 0.263) + Fbx
    alam_aV[ultraviolet] = ax + bx/Rv
    
    #-- far UV
    fuv = 8.0<=all_x
    x = all_x[fuv]
    ax = -1.073 - 0.628*(x-8) + 0.137*(x-8)**2 - 0.070*(x-8)**3
    bx = 13.670 + 4.257*(x-8) - 0.420*(x-8)**2 + 0.374*(x-8)**3
    alam_aV[fuv] = ax + bx/Rv
    
    return wave,alam_aV


def seaton1979(Rv=3.1,wave=None,**kwargs):
    """
    Extinction curve from Seaton, 1979.
    
    This function returns A(lambda)/A(V).
    
    To get A(lambda)/E(B-V), multiply the return value with Rv (A(V)=Rv*E(B-V))
    
    Extra kwags are to catch unwanted keyword arguments.
    
    @param Rv: Rv
    @type Rv: float
    @param wave: wavelengths to compute the curve on
    @type wave: ndarray
    @return: wavelengths (A), A(lambda)/Av
    @rtype: (ndarray,ndarray)
    """
    if wave is None: wave = np.r_[1000.:10000.:10]
    
    all_x = 1e4/(wave)
    alam_aV = np.zeros_like(all_x)
    
    #-- far infrared
    x_ = np.r_[1.0:2.8:0.1]
    X_ = np.array([1.36,1.44,1.84,2.04,2.24,2.44,2.66,2.88,3.14,3.36,3.56,3.77,3.96,4.15,4.26,4.40,4.52,4.64])
    fir = all_x<=2.7
    alam_aV[fir] = np.interp(all_x[fir][::-1],x_,X_,left=0)[::-1]
    
    #-- infrared
    infrared = (2.70<=all_x) & (all_x<3.65)
    x = all_x[infrared]
    alam_aV[infrared] = 1.56 + 1.048*x + 1.01 / ( (x-4.60)**2 + 0.280)
    
    #-- optical
    optical = (3.65<=all_x) & (all_x<7.14)
    x = all_x[optical]
    alam_aV[optical] = 2.29 + 0.848*x + 1.01 / ( (x-4.60)**2 + 0.280)
    
    #-- ultraviolet
    ultraviolet = (7.14<=all_x) & (all_x<=10)
    x = all_x[ultraviolet]
    alam_aV[ultraviolet] = 16.17 - 3.20*x + 0.2975*x**2
    
    return wave,alam_aV/Rv

#}

if __name__=="__main__":
    import doctest
    import pylab as pl
    doctest.testmod()
    pl.show()